import importlib
import os
import shutil
import warnings
from pathlib import Path
from typing import Any

import narwhals as nw
from pydantic import AliasChoices
from pydantic import Field
from pydantic import computed_field
from pydantic import field_serializer
from pydantic import model_validator

from laktory._logger import get_logger
from laktory._settings import settings
from laktory.enums import STREAMING_BACKENDS
from laktory.enums import DataFrameBackends
from laktory.exceptions import DataQualityExpectationsNotSupported
from laktory.models.basemodel import BaseModel
from laktory.models.dataframe.dataframetransformer import DataFrameTransformer
from laktory.models.dataquality.expectation import DataQualityExpectation
from laktory.models.datasinks import DataSinksUnion
from laktory.models.datasinks import TableDataSink
from laktory.models.datasources import BaseDataSource
from laktory.models.datasources import DataSourcesUnion
from laktory.models.datasources import PipelineNodeDataSource
from laktory.models.datasources import TableDataSource
from laktory.models.pipelinechild import PipelineChild
from laktory.typing import AnyFrame

logger = get_logger(__name__)


class PipelineNode(BaseModel, PipelineChild):
    """
    Pipeline base component generating a DataFrame by reading a data source and
    applying a transformer (chain of dataframe transformations). Optional
    output to a data sink.

    Examples
    --------
    A node reading stock prices data from a CSV file and writing a DataFrame
    as a parquet file.
    ```py
    import io

    import laktory as lk

    node_yaml = '''
        name: brz_stock_prices
        source:
          path: "./events/stock_prices/"
          format: JSON
        sinks:
        - path: ./tables/brz_stock_prices/
          format: PARQUET
    '''

    node = lk.models.PipelineNode.model_validate_yaml(io.StringIO(node_yaml))

    # node.execute()
    ```

    A node reading stock prices from an upstream node and writing a DataFrame
    to a data table.
    ```py
    import io

    import laktory as lk

    node_yaml = '''
        name: slv_stock_prices
        source:
          node_name: brz_stock_prices
        sinks:
        - schema_name: finance
          table_name: slv_stock_prices
        transformer:
          nodes:
          - expr: |
                SELECT
                  data.created_at AS created_at,
                  data.symbol AS symbol,
                  data.open AS open,
                  data.close AS close,
                  data.high AS high,
                  data.low AS low,
                  data.volume AS volume
                FROM
                  {df}
          - func_name: drop_duplicates
            func_kwargs:
              subset:
                - symbol
                - timestamp
    '''

    node = lk.models.PipelineNode.model_validate_yaml(io.StringIO(node_yaml))

    # node.execute()
    ```
    References
    ----------
    * [Data Pipeline](https://www.laktory.ai/concepts/pipeline/)
    """

    dlt_template: str | None = Field(
        "DEFAULT",
        description="Specify which template (notebook) to use when Databricks pipeline is selected as the orchestrator.",
    )
    comment: str = Field(
        None,
        description="Comment for the associated table or view",
    )
    expectations: list[DataQualityExpectation] = Field(
        [],
        description="List of data expectations. Can trigger warnings, drop invalid records or fail a pipeline.",
    )
    expectations_checkpoint_path_: str | Path = Field(
        None,
        description="Path to which the checkpoint file for which expectations on a streaming dataframe should be written.",
        validation_alias=AliasChoices(
            "expectations_checkpoint_path", "expectations_checkpoint_path_"
        ),
        exclude=True,
    )
    name: str = Field(..., description="Name given to the node.")
    primary_keys: list[str] = Field(
        None,
        description="""
        A list of column names that uniquely identify each row in the
        DataFrame. These columns are used to:
            - Document the uniqueness constraints of the node's output data.
            - Define the default primary keys for sinks CDC merge operations
            - Referenced in expectations and unit tests.
        While optional, specifying `primary_keys` helps enforce data integrity
        and ensures that downstream operations, such as deduplication, are
        consistent and reliable.
        """,
    )
    source: DataSourcesUnion | None = Field(
        None, description="Definition of the data source(s)"
    )
    sinks: list[DataSinksUnion] = Field(
        None,
        description="Definition of the data sink(s). Set `is_quarantine` to True to store node quarantine DataFrame.",
    )
    time_column: str | None = Field(
        None,
        description="""
        The name of the column that represents the timestamp or temporal dimension in the DataFrame. This column is 
        used to:
            - Document the time-based ordering and filtering semantics of the node’s output data.
            - Enable time-aware operations such as point-in-time joins, incremental processing, and time-series analysis.
            - Serve as a reference in expectations, unit tests, and feature engineering workflows.
        While optional, specifying time_column helps ensure consistency in time-based logic and improves the 
        reliability of downstream operations that depend on temporal alignment.        
        """,
    )
    transformer: DataFrameTransformer = Field(
        None,
        description="Data transformations applied between the source and the sink(s).",
    )
    root_path_: str | Path = Field(
        None,
        description="Location of the pipeline node root used to store logs, metrics and checkpoints.",
        validation_alias=AliasChoices("root_path", "root_path_"),
        exclude=True,
    )
    _stage_df: Any = None
    _output_df: Any = None
    _quarantine_df: Any = None

    @model_validator(mode="after")
    def push_primary_keys(self) -> Any:
        # Assign primary keys
        if self.primary_keys and self.sinks:
            for s in self.sinks:
                if s.is_cdc:
                    if s.merge_cdc_options.primary_keys is None:
                        s.merge_cdc_options.primary_keys = self.primary_keys

        return self

    @model_validator(mode="after")
    def validate_expectations(self):
        if self.source and self.source.as_stream:
            # Expectations type
            for e in self.expectations:
                if not e.is_streaming_compatible:
                    raise DataQualityExpectationsNotSupported(e, self)

            if self.expectations and self.expectations_checkpoint_path is None:
                warnings.warn(
                    f"Node '{self.name}' requires `expectations_checkpoint_location` specified unless Databricks pipeline is selected as an orchestrator and expectations are compatible with Declarative Pipelines."
                )

        return self

    @model_validator(mode="after")
    def validate_view(self):
        if not self.is_view:
            return self

        # Validate Source
        if self.source:
            if not (
                isinstance(self.source, TableDataSource)
                or isinstance(self.source, PipelineNodeDataSource)
            ):
                raise ValueError(
                    "VIEW sink only supports Table or Pipeline Node with Table sink Data Source"
                )

            if self.source.as_stream:
                raise ValueError("VIEW sink does not support stream read.")

        # Validate Sinks
        m = f"node '{self.name}': "
        for s in self.sinks:
            if s.table_type != "VIEW":
                raise ValueError(
                    f"{m}If one pipeline node sink is a VIEW, all of them must be a view."
                )

        # Validate Expectations
        if self.expectations:
            raise ValueError(f"{m}Expectations not supported for a view sink.")

        return self

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def children_names(self):
        return [
            "source",
            "data_sources",
            "expectations",
            "transformer",
            "sinks",
        ]

    # ----------------------------------------------------------------------- #
    # Orchestrator                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def is_orchestrator_dlt(self) -> bool:
        """If `True`, pipeline node is used in the context of a DLT pipeline"""
        pl = self.parent_pipeline
        if pl:
            return pl.is_orchestrator_dlt
        return False

    @property
    def is_dlt_execute(self) -> bool:
        if not self.is_orchestrator_dlt:
            return False
        from laktory import is_dlt_execute

        return is_dlt_execute()

    # ----------------------------------------------------------------------- #
    # Paths                                                                   #
    # ----------------------------------------------------------------------- #

    @computed_field(description="root_path")
    @property
    def root_path(self) -> Path:
        if self.root_path_:
            return Path(self.root_path_)

        pl = self.parent_pipeline
        if pl and pl.root_path:
            return pl.root_path / self.name

        return Path(settings.laktory_root) / self.name

    @field_serializer("root_path", "expectations_checkpoint_path", when_used="json")
    def serialize_path(self, value: Path) -> str:
        return value.as_posix()

    # ----------------------------------------------------------------------- #
    # Outputs and Sinks                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def is_view(self) -> bool:
        if not self.sinks:
            return False
        is_view = False
        for s in self.sinks:
            if isinstance(s, TableDataSink) and s.table_type == "VIEW":
                is_view = True
                break
        return is_view

    @property
    def stage_df(self) -> AnyFrame:
        """
        Dataframe resulting from reading source and applying transformer, before data quality checks are applied.
        """
        return self._stage_df

    @property
    def output_df(self) -> AnyFrame:
        """
        Dataframe resulting from reading source, applying transformer and dropping rows not meeting data quality
        expectations.
        """
        return self._output_df

    @property
    def quarantine_df(self) -> AnyFrame:
        """
        DataFrame storing `stage_df` rows not meeting data quality expectations.
        """
        return self._quarantine_df

    @property
    def output_sinks(self) -> list[DataSinksUnion]:
        """List of sinks writing the output DataFrame"""
        sinks = []
        if self.sinks:
            for s in self.sinks:
                if not s.is_quarantine:
                    sinks += [s]
        return sinks

    @property
    def quarantine_sinks(self) -> list[DataSinksUnion]:
        """List of sinks writing the quarantine DataFrame"""
        sinks = []
        if self.sinks:
            for s in self.sinks:
                if s.is_quarantine:
                    sinks += [s]
        return sinks

    @property
    def all_sinks(self):
        """List of all sinks (output and quarantine)."""
        return self.output_sinks + self.quarantine_sinks

    @property
    def sinks_count(self) -> int:
        """Total number of sinks."""
        return len(self.all_sinks)

    @property
    def has_output_sinks(self) -> bool:
        """`True` if node has at least one output sink."""
        return len(self.output_sinks) > 0

    # @property
    # def has_quarantine_sinks(self) -> bool:
    #     """`True` if node has at least one quarantine sink."""
    #     return len(self.quarantine_sinks) > 0

    @property
    def has_sinks(self) -> bool:
        """`True` if node has at least one sink."""
        return self.sinks_count > 0

    @property
    def primary_sink(self) -> DataSinksUnion | None:
        """Primary output sink used as a source for downstream nodes."""
        if not self.has_output_sinks:
            return None

        return self.output_sinks[0]

    # ----------------------------------------------------------------------- #
    # Expectations                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def dlt_warning_expectations(self) -> dict[str, str]:
        expectations = {}
        for e in self.expectations:
            if e.is_dlt_compatible and e.action == "WARN":
                expectations[e.name] = e.expr.expr
        return expectations

    @property
    def dlt_drop_expectations(self) -> dict[str, str]:
        expectations = {}
        for e in self.expectations:
            if e.is_dlt_compatible and e.action in ["DROP", "QUARANTINE"]:
                expectations[e.name] = e.expr.expr
        return expectations

    @property
    def dlt_fail_expectations(self) -> dict[str, str]:
        expectations = {}
        for e in self.expectations:
            if e.is_dlt_compatible and e.action == "FAIL":
                expectations[e.name] = e.expr.expr
        return expectations

    @computed_field(description="expectations_checkpoint_path")
    @property
    def expectations_checkpoint_path(self) -> Path | None:
        if self.expectations_checkpoint_path_:
            return Path(self.expectations_checkpoint_path_)

        if self.root_path:
            return Path(self.root_path) / "checkpoints/expectations"

        return None

    @property
    def checks(self):
        return [e.check for e in self.expectations]

    # ----------------------------------------------------------------------- #
    # Upstream Nodes                                                          #
    # ----------------------------------------------------------------------- #

    @property
    def upstream_node_names(self) -> list[str]:
        """Pipeline node names required to execute current node."""
        from laktory.models.datasources.pipelinenodedatasource import (
            PipelineNodeDataSource,
        )

        names = []

        if isinstance(self.source, PipelineNodeDataSource):
            names += [self.source.node_name]

        if self.transformer:
            names += self.transformer.upstream_node_names

        if self.sinks:
            for s in self.sinks:
                names += s.upstream_node_names

        return names

    # ----------------------------------------------------------------------- #
    # Data Sources                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def data_sources(self) -> list[BaseDataSource]:
        """Get all sources feeding the pipeline node"""
        sources = []

        if isinstance(self.source, BaseDataSource):
            sources += [self.source]

        if self.transformer:
            sources += self.transformer.data_sources

        if self.sinks:
            for s in self.sinks:
                sources += s.data_sources

        return sources

    # ----------------------------------------------------------------------- #
    # Execution                                                               #
    # ----------------------------------------------------------------------- #

    def purge(self):
        logger.info(f"Purging pipeline node {self.name}")

        if self.has_sinks:
            for s in self.sinks:
                s.purge()

        if self.expectations_checkpoint_path:
            if os.path.exists(self.expectations_checkpoint_path):
                logger.info(
                    f"Deleting expectations checkpoint at {self.expectations_checkpoint_path}",
                )
                shutil.rmtree(self.expectations_checkpoint_path)

            # Try with DBFS
            # If spark is not used, dbfs is most likely not used
            if self.dataframe_backend != DataFrameBackends.PYSPARK:
                return

            # Check if a workspace client can be instantiated
            try:
                from databricks.sdk import WorkspaceClient
                from databricks.sdk.errors import ResourceDoesNotExist

                w = WorkspaceClient()

            except (
                ModuleNotFoundError,  # SDK not installed
                ImportError,  # SDK with different version / API
                ValueError,  # client not configure (would never happen in a notebook)
            ):
                return

            # Format path for DBFS
            _path = self.expectations_checkpoint_path.as_posix()
            if not _path.startswith("/") and not _path.startswith("dbfs:"):
                _path = "/" + _path
            if _path.startswith("/dbfs/"):
                _path = _path.replace("/dbfs/", "dbfs:/")
            if not _path.startswith("dbfs:"):
                _path = "dbfs:" + _path

            # Check Status
            try:
                w.dbfs.get_status(_path)
            except ResourceDoesNotExist:
                logger.info(
                    f"Expectation checkpoint at {_path} does not exist. Skipping.",
                )
                return

            logger.info(
                f"Deleting expectation checkpoint at {_path}.",
            )
            w.dbfs.delete(_path, recursive=True)

    def execute(
        self,
        apply_transformer: bool = True,
        write_sinks: bool = True,
        full_refresh: bool = False,
        named_dfs: dict[str, AnyFrame] = None,
        update_tables_metadata: bool = True,
    ) -> AnyFrame:
        """
        Execute pipeline node by:

        - Reading the source
        - Applying the user defined (and layer-specific if applicable) transformations
        - Checking expectations
        - Writing the sinks

        Parameters
        ----------
        apply_transformer:
            Flag to apply transformer in the execution
        write_sinks:
            Flag to include writing sink in the execution
        full_refresh:
            If `True` dataframe will be completely re-processed by deleting
            existing data and checkpoint before processing.
        named_dfs:
            Named DataFrame passed to transformer nodes
        update_tables_metadata:
            Update tables metadata

        Returns
        -------
        :
            output Spark DataFrame
        """
        logger.info(f"Executing pipeline node {self.name}")

        # Install dependencies
        pl = self.parent_pipeline
        if pl and not pl._imports_imported:
            for package_name in pl._imports:
                try:
                    logger.info(f"Importing {package_name}")
                    importlib.import_module(package_name)
                except ModuleNotFoundError:
                    logger.info(f"Importing {package_name} failed.")
            pl._imports_imported = True

        # Parse DLT
        if self.is_orchestrator_dlt:
            logger.info("DLT orchestrator selected. Sinks writing will be skipped.")
            write_sinks = False
            full_refresh = False

        # Refresh
        if full_refresh:
            self.purge()

        # Read Source
        self._stage_df = None
        if self.source:
            self._stage_df = self.source.read()

        # Apply transformer
        if named_dfs is None:
            named_dfs = {}
        if apply_transformer and self.transformer:
            self._stage_df = self.transformer.execute(
                self._stage_df, named_dfs=named_dfs
            )

        # Check expectations
        self._output_df = self._stage_df
        self._quarantine_df = None
        self.check_expectations()

        # Output and Quarantine to Sinks
        if write_sinks:
            for s in self.output_sinks:
                if self.is_view:
                    s.write()
                    self._output_df = s.as_source().read()
                else:
                    s.write(self._output_df, full_refresh=full_refresh)

            if self._quarantine_df is not None:
                for s in self.quarantine_sinks:
                    s.write(self._quarantine_df, full_refresh=full_refresh)

        # Update tables metadata
        if update_tables_metadata and not self.is_dlt_execute:
            for s in self.all_sinks:
                if s.metadata:
                    s.metadata.execute()

        return self._output_df

    def check_expectations(self):
        """
        Check expectations, raise errors, warnings where required and build
        filtered and quarantine DataFrames.

        Some actions have to be disabled when selected orchestrator is
        Databricks DLT:

        * Raising error on Failure when expectation is supported by DLT
        * Dropping rows when expectation is supported by DLT
        """

        # Data Quality Checks
        qfilter = None  # Quarantine filter
        kfilter = None  # Keep filter
        if self._stage_df is None:
            # Node without source or transformer
            return
        is_streaming = getattr(nw.to_native(self._stage_df), "isStreaming", False)
        if not self.expectations:
            return

        def _batch_check(df, node):
            for e in node.expectations:
                # Run Check: this only warn or raise exceptions.
                if not e.is_dlt_managed:
                    e.run_check(
                        df,
                        raise_or_warn=True,
                        node=node,
                    )

        def _stream_check(batch_df, batch_id, node):
            _batch_check(
                batch_df,
                node,
            )

        logger.info("Checking Data Quality Expectations")

        if not is_streaming:
            _batch_check(
                self._stage_df,
                self,
            )

        else:
            skip = False

            if self.is_dlt_execute:
                names = []
                for e in self.expectations:
                    if not e.is_dlt_compatible:
                        names += [e.name]
                if names:
                    raise TypeError(
                        f"Expectations {names} are not natively supported by DLT and can't be computed on a streaming DataFrame with DLT executor."
                    )

                skip = True

            backend = DataFrameBackends.from_df(self._stage_df)
            if backend not in STREAMING_BACKENDS:
                raise TypeError(
                    f"DataFrame backend {backend} is not supported for streaming operations"
                )

            if self.expectations_checkpoint_path is None:
                raise ValueError(
                    f"Expectations Checkpoint not specified for node '{self.name}'"
                )

            # TODO: Refactor for backend other than spark
            if not skip:
                query = (
                    self._stage_df.to_native()
                    .writeStream.foreachBatch(
                        lambda batch_df, batch_id: _stream_check(
                            nw.from_native(batch_df), batch_id, self
                        )
                    )
                    .trigger(availableNow=True)
                    .options(
                        checkpointLocation=self.expectations_checkpoint_path,
                    )
                    .start()
                )
                query.awaitTermination()

        # Build Filters
        for e in self.expectations:
            # Update Keep Filter
            if not e.is_dlt_managed:
                _filter = e.keep_filter
                if _filter is not None:
                    if kfilter is None:
                        kfilter = _filter
                    else:
                        kfilter = kfilter & _filter

            # Update Quarantine Filter
            _filter = e.quarantine_filter
            if _filter is not None:
                if qfilter is None:
                    qfilter = _filter
                else:
                    qfilter = qfilter & _filter

        if qfilter is not None:
            logger.info("Building quarantine DataFrame")
            self._quarantine_df = self._stage_df.filter(qfilter)
        else:
            self._quarantine_df = self._stage_df  # .filter("False")

        if kfilter is not None:
            logger.info("Dropping invalid rows")
            self._output_df = self._stage_df.filter(kfilter)
        else:
            self._output_df = self._stage_df



import logging
import sys
from datetime import datetime
from zoneinfo import ZoneInfo

from ._settings import settings

# --------------------------------------------------------------------------- #
# Custom Formatter                                                            #
# --------------------------------------------------------------------------- #


class LaktoryFormatter(logging.Formatter):
    info_format = "%(asctime)s [laktory] %(message)s"
    default_format = "%(asctime)s [%(name)s] %(levelname)s | %(message)s"

    def __init__(self, fmt=default_format):
        super().__init__(fmt=fmt, datefmt="%Y-%m-%d %H:%M:%S")

    def formatTime(self, record, datefmt=None):
        # Set timestamp to UTC
        utc_time = datetime.fromtimestamp(record.created, tz=ZoneInfo("UTC"))
        return utc_time.strftime(datefmt or self.default_time_format)

    def format(self, record):
        # Use info-specific format, otherwise use default
        if record.levelno == logging.INFO:
            self._style._fmt = self.info_format
        else:
            self._style._fmt = self.default_format
        return super().format(record)


# --------------------------------------------------------------------------- #
# Get logger                                                                  #
# --------------------------------------------------------------------------- #

# Stream Handler
stream_handler = logging.StreamHandler(sys.stdout)
# format = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
# stream_handler.setFormatter(format)
stream_handler.setFormatter(LaktoryFormatter())


def get_logger(name, stream=True):
    # Change logger class
    # TODO: Place Holder if we need to customize logger class
    # logging.setLoggerClass(LaktoryLogger)

    # Get logger
    logger = logging.getLogger(name)

    # Set level
    logger.setLevel(settings.log_level)

    # Set options
    _options = {
        "stream": stream,
    }

    # Set states
    _initialized = hasattr(logger, "_initialized")
    _options_changed = False
    if _initialized:
        _options_changed = logger._options != _options

    # Update handlers
    if not _initialized or _options_changed:
        handlers = list(logger.handlers)
        for handler in handlers:
            logger.removeHandler(handler)

        if stream:
            logger.addHandler(stream_handler)

    # Reset logger class
    logging.setLoggerClass(logging.Logger)

    # Update state
    logger._initialized = True
    logger._options = _options

    return logger



from pathlib import Path

from pydantic import Field

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.notebook import Notebook
from laktory.models.resources.databricks.workspacefile import WorkspaceFile
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource

logger = get_logger(__name__)


class WorkspaceTree(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Workspace Tree (collections of directories, notebooks and
    workspace files)

    Examples
    --------
    ```py
    from laktory import models

    tree = models.resources.databricks.WorkspaceTree(
        source="./source/",
    )
    print(tree)
    '''
    resource_name_=None options=ResourceOptions(variables={}, is_enabled=True, depends_on=[], provider=None, ignore_changes=None, aliases=None, delete_before_replace=True, import_=None, parent=None, replace_on_changes=None, moved_from=None) lookup_existing=None variables={} access_controls=[] path=None source='./source/'
    '''
    ```
    """

    access_controls: list[AccessControl] = Field([], description="Access controls list")
    path: str = Field(
        None,
        description="Workspace filepath for the tree. If not specified, workspace laktory root is used.",
    )
    source: str = Field(
        ...,
        description="Path to directory on local filesystem.",
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def self_as_core_resources(self) -> bool:
        return False

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        resources = []

        # Get file paths
        source = Path(self.source)
        root = source.resolve()
        filepaths = []
        for filepath in root.rglob("*"):
            if filepath.is_dir():
                continue
            if filepath.name.startswith("."):
                continue
            filepaths += [filepath]
        filepaths.sort()

        # Create resources
        for filepath in filepaths:
            # Check if notebook
            is_notebook = filepath.suffix == ".ipynb"
            if filepath.suffix == ".py":
                content = filepath.read_text()
                if "# Databricks notebook source" in content:
                    is_notebook = True

            # Set source (local file system)
            if source.is_absolute():
                _source = str(filepath)
            else:
                _source = filepath.relative_to(root.parent)

            # Set path (Databricks / unix file system)
            dirpath = str(filepath.parent).replace(str(root), "")
            if self.path:
                if dirpath.startswith("/"):
                    dirpath = dirpath[1:]
                kwargs = {
                    "path": (Path(self.path) / dirpath / filepath.name).as_posix()
                }
            else:
                kwargs = {"dirpath": Path(dirpath).as_posix()}

            # Set access controls
            kwargs["access_controls"] = self.access_controls

            if is_notebook:
                r = Notebook(source=str(_source), **kwargs)
            else:
                r = WorkspaceFile(source=str(_source), **kwargs)

            resources += [r]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str | None:
        return None

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str | None:
        return None



from typing import TYPE_CHECKING
from typing import Any

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field

if TYPE_CHECKING:
    from databricks.sdk import WorkspaceClient


class DispatcherRunner(BaseModel):
    """
    Base runner for jobs, pipelines and other executables.
    """

    model_config = ConfigDict(extra="forbid")
    name: str = Field(None, description="Name of the resource attached to the runner")
    id: str = Field(None, description="ID of the deployed resource")
    dispatcher: Any = Field(
        default=None, exclude=True, description="Dispatcher managing the runs"
    )

    @property
    def wc(self) -> "WorkspaceClient":
        """Databricks Workspace Client"""
        return self.dispatcher.wc

    def get_id(self) -> str:
        raise NotImplementedError()

    def run(self, wait=True):
        raise NotImplementedError()



from typing import Any
from typing import Literal
from typing import Union

import narwhals as nw
from pydantic import Field
from pydantic import field_validator
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel

logger = get_logger(__name__)

NAMES = [
    "Array",
    "Boolean",
    "Categorical",
    "Date",
    "Datetime",
    "Decimal",
    "Duration",
    "Enum",
    "Float32",
    "Float64",
    "Int128",
    "Int16",
    "Int32",
    "Int64",
    "Int8",
    "List",
    "Object",
    "String",
    "Struct",
    "UInt128",
    "UInt16",
    "UInt32",
    "UInt64",
    "UInt8",
    "Unknown",
]

ALIASES = {
    "bigint": "Int64",
    "int": "Int32",
    "integer": "Int32",
    "smallint": "Int16",
    "short": "Int16",
    "tinyint": "Int8",
    "byte": "Int8",
    "float": "Float32",
    "double": "Float64",
    "str": "String",
    "varchar": "String",
    "text": "String",
    "bool": "Boolean",
    "obj": "Object",
    "binary": "Object",
    "varbinary": "Object",
    "category": "Categorical",
    "timestamp": "Datetime",
}


__all__ = ["DType", "DField"] + NAMES


class DField(BaseModel):
    """
    Data Field definition
    """

    name: str = Field(..., description="Field name")
    dtype: Union[str, "DType"] = Field(..., description="Field data type")

    @field_validator("dtype", mode="before")
    def update_dtype(cls, dtype: Any) -> Any:
        if isinstance(dtype, str):
            dtype = DType(name=dtype)
        return dtype


class DType(BaseModel):
    """
    Generic data type class.

    Examples
    --------
    ```
    from laktory import models

    # Int32
    dtype = models.DType(name="Int32")
    print(dtype)

    # List of string
    dtype = models.DType(name="list", inner="str")
    print(dtype)

    # Structure
    dtype = models.DType(
        name="struct", fields={"x": {"name": "list", "inner": "double"}, "y": dtype}
    )
    print(dtype)
    ```
    """

    name: str = Field(..., description="Data type name.")
    inner: Union[str, "DType"] = Field(
        None, description="Data type for sub-elements for `Array` or `List` types."
    )
    fields: list[DField] = Field(
        None, description="Definition of fields for `Struct` type."
    )
    shape: int | list[int] = Field(
        None, description="Definition of shape for `Array` type."
    )
    category: Literal["NUMERIC", "STRING", "STRUCT"] = Field(
        "NUMERIC", description="Data type category"
    )

    @field_validator("name", mode="before")
    @classmethod
    def validate_name(cls, name: Any) -> Any:
        # From Names
        for _name in NAMES:
            if name.lower() == _name.lower():
                return _name

        # From Aliases
        _name = ALIASES.get(name.lower(), None)
        if _name:
            logger.warn(f"Data type '{name}' should be replaced with '{_name}'")
            return _name

        raise ValueError(f"'{name}' type is not supported.")

    @model_validator(mode="after")
    def complex_types(self) -> Any:
        if self.name in ["Array", "List"] and self.inner is None:
            raise ValueError(f"'{self.name}' type requires `inner` value.")
        if self.name in ["Array"] and self.shape is None:
            raise ValueError(f"'{self.name}' type requires `shape` value.")
        if self.name in ["Struct"] and self.fields is None:
            raise ValueError(f"'{self.name}' type requires `fields` value.")
        return self

    @field_validator("inner", mode="before")
    def update_inner(cls, v: Any) -> Any:
        if isinstance(v, str):
            return DType(name=v)
        if isinstance(v, DType) and type(v) is not DType:
            """Required to prevent serialization issue with pydantic"""
            dump = v.model_dump(exclude_unset=True)  # Required to avoid None in inner
            return DType(**dump)

        return v

    # @field_validator("fields", mode="before")
    # def update_fields(cls, fields: Any) -> Any:
    #     if fields is None:
    #         return
    #
    #     for k, v in fields.items():
    #         if isinstance(v, str):
    #             fields[k] = DType(name=v)
    #
    #         elif isinstance(v, DType) and type(v) is not DType:
    #             """Required to prevent serialization issue with pydantic"""
    #             dump = v.model_dump(exclude_unset=True)
    #             fields[k] = DType(**dump)
    #
    #     return fields

    def to_generic(self):
        return DType(**self.model_dump(exclude_unset=True))

    def to_narwhals(self):
        """Get equivalent Narwhals data type"""
        nw_dtypes = nw.dtypes
        _type = self.name

        # Complex types
        if _type in "Array":
            return nw_dtypes.Array(inner=self.inner.to_narwhals(), shape=self.shape)

        if _type == "List":
            return nw_dtypes.List(inner=self.inner.to_narwhals())

        if _type == "Struct":
            fields = []
            for field in self.fields:
                fields += [nw.Field(name=field.name, dtype=field.dtype.to_narwhals())]
            return nw_dtypes.Struct(fields)

        if hasattr(nw_dtypes, _type):
            return getattr(nw_dtypes, _type)

        # Not Found
        raise ValueError(f"Data type with name '{self.name}' is not supported")

    def to_spark(self):
        """Get equivalent Spark data type"""
        import pyspark.sql.types as T
        from narwhals._spark_like.utils import narwhals_to_native_dtype

        from laktory import get_spark_session

        spark = get_spark_session()
        return narwhals_to_native_dtype(
            dtype=self.to_narwhals(),
            version=nw._utils.Version.MAIN,
            spark_types=T,
            session=spark,
        )

    def to_polars(self):
        """Get equivalent Polars data type"""
        from narwhals._polars.utils import narwhals_to_native_dtype

        return narwhals_to_native_dtype(
            dtype=self.to_narwhals(),
            version=nw._utils.Version.MAIN,
        )

    def to_string(self):
        return str(self.to_polars())


class SpecificDType(DType):
    @model_validator(mode="after")
    def set_name(self) -> Any:
        """Force dump of `name` property to allow de-serialization from DType model"""
        self.model_fields_set.add("name")
        return self


# Complex types
class Array(SpecificDType):
    name: str = Field("Array", frozen=True)
    inner: str | DType
    shape: int | list[int]


class List(SpecificDType):
    name: str = Field("List", frozen=True)
    inner: str | DType


class Struct(SpecificDType):
    name: str = Field("Struct", frozen=True)
    fields: list[DField]


# Integer types
class Decimal(SpecificDType):
    name: str = Field("Decimal", frozen=True)


class Int128(SpecificDType):
    name: str = Field("Int128", frozen=True)


class Int64(SpecificDType):
    name: str = Field("Int64", frozen=True)


class Int32(SpecificDType):
    name: str = Field("Int32", frozen=True)


class Int16(SpecificDType):
    name: str = Field("Int16", frozen=True)


class Int8(SpecificDType):
    name: str = Field("Int8", frozen=True)


# Unsigned Integer types
class UInt128(SpecificDType):
    name: str = Field("UInt128", frozen=True)


class UInt64(SpecificDType):
    name: str = Field("UInt64", frozen=True)


class UInt32(SpecificDType):
    name: str = Field("UInt32", frozen=True)


class UInt16(SpecificDType):
    name: str = Field("UInt16", frozen=True)


class UInt8(SpecificDType):
    name: str = Field("UInt8", frozen=True)


# Floating-point types
class Float32(SpecificDType):
    name: str = Field("Float32", frozen=True)


class Float64(SpecificDType):
    name: str = Field("Float64", frozen=True)


# String & Boolean
class String(SpecificDType):
    name: str = Field("String", frozen=True)


class Boolean(SpecificDType):
    name: str = Field("Boolean", frozen=True)


# Special types
class Object(SpecificDType):
    name: str = Field("Object", frozen=True)


class Categorical(SpecificDType):
    name: str = Field("Categorical", frozen=True)


class Enum(SpecificDType):
    name: str = Field("Enum", frozen=True)


# Date & Time types
class Datetime(SpecificDType):
    name: str = Field("Datetime", frozen=True)


class Duration(SpecificDType):
    name: str = Field("Duration", frozen=True)


class Date(SpecificDType):
    name: str = Field("Date", frozen=True)


# Unknown
class Unknown(SpecificDType):
    name: str = Field("Unknown", frozen=True)


DType.model_rebuild()
DField.model_rebuild()



from .dataframecolumn import DataFrameColumn
from .dataframecolumnexpr import DataFrameColumnExpr
from .dataframeexpr import DataFrameExpr
from .dataframemethod import DataFrameMethod
from .dataframeschema import DataFrameSchema
from .dataframetransformer import DataFrameTransformer



import json
from typing import Any
from typing import Literal
from typing import Union

from pydantic import ConfigDict
from pydantic import Field
from pydantic import field_validator
from pydantic import model_validator

from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.cluster import Cluster
from laktory.models.resources.databricks.cluster import ClusterLibrary
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class JobJobClusterNewCluster(Cluster):
    """
    Job Cluster. Same attributes as `laktory.models.Cluster`, except for

    * `access_controls`
    * `is_pinned`
    * `libraries`
    * `name`
    * `no_wait`

    that are not allowed.
    """

    access_controls: list[Any] = Field(None, exclude=True)
    is_pinned: bool = Field(None, exclude=True)
    libraries: list[Any] = Field(None, exclude=True)
    name: str = Field(None, exclude=True)
    no_wait: bool = Field(None, exclude=True)

    @model_validator(mode="after")
    def excluded_fields(self) -> Any:
        for f in [
            "access_controls",
            "is_pinned",
            "libraries",
            "name",
            "no_wait",
        ]:
            if getattr(self, f, None) not in [None, [], {}]:
                raise ValueError(f"Field {f} should be null")

        return self

    @property
    def resource_key(self) -> str:
        return ""


class JobJobCluster(BaseModel):
    job_cluster_key: str = Field(
        ...,
        description="Identifier that can be referenced in task block, so that cluster is shared between tasks",
    )
    new_cluster: JobJobClusterNewCluster = Field(
        ...,
        description="""
    Block with almost the same set of parameters as for databricks.Cluster resource, except following (check the REST
    API documentation for full list of supported parameters):
    """,
    )


class JobContinuous(BaseModel):
    pause_status: Union[Literal["PAUSED", "UNPAUSED"], str] = Field(
        None,
        description="""
    Indicate whether this continuous job is paused or not. When the pause_status field is omitted in the block,
    the server will default to using `UNPAUSED` as a value for pause_status.
    """,
    )


class JobEmailNotifications(BaseModel):
    no_alert_for_skipped_runs: bool = Field(
        None,
        description="""
    If `True`, don't send alert for skipped runs. (It's recommended to use the corresponding setting in the
    notification_settings configuration block).
    """,
    )
    on_duration_warning_threshold_exceededs: list[str] = Field(
        None,
        description="""
    List of emails to notify when the duration of a run exceeds the threshold specified by the RUN_DURATION_SECONDS
    metric in the health block.
    """,
    )
    on_failures: list[str] = Field(
        None, description="List of emails to notify when the run fails."
    )
    on_starts: list[str] = Field(
        None, description="List of emails to notify when the run starts."
    )
    on_successes: list[str] = Field(
        None,
        description="List of emails to notify when the run completes successfully.",
    )

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "on_failures": "on_failure",
            "on_starts": "on_start",
            "on_successes": "on_success",
            "on_duration_warning_threshold_exceededs": "on_duration_warning_threshold_exceeded",
        }


class JobEnvironmentSpec(BaseModel):
    client: str = Field(..., description="client version used by the environment")
    dependencies: list[str] = Field(
        None,
        description="""
    List of pip dependencies, as supported by the version of pip in this environment. Each dependency is a pip 
    requirement file line. See 
    [API docs](https://docs.databricks.com/api/workspace/jobs/create#environments-spec-dependencies)
    for more information.
    """,
    )


class JobEnvironment(BaseModel):
    environment_key: str = Field(
        ...,
        description="""
        An unique identifier of the Environment. It will be referenced from environment_key attribute of corresponding 
        task.
        """,
    )
    spec: JobEnvironmentSpec = Field(None, description="")


class JobGitSourceGitSnapshot(BaseModel):
    used_commit: str = Field(..., description="")


class JobGitSourceJobSource(BaseModel):
    import_from_git_branch: str = Field(..., description="")
    job_config_path: str = Field(..., description="")
    dirty_state: str = Field(None, description="")


class JobGitSource(BaseModel):
    url: str = Field(..., description="URL of the Git repository to use.")
    branch: str = Field(
        None,
        description="Name of the Git branch to use. Conflicts with `tag` and `commit`.",
    )
    commit: str = Field(
        None,
        description="Hash of Git commit to use. Conflicts with `branch` and `tag`.",
    )
    git_snapshot: JobGitSourceGitSnapshot = Field(
        None, description="Git snapshot specifications"
    )
    job_source: JobGitSourceJobSource = Field(
        None, description="Job source specifications"
    )
    provider: str = Field(
        ...,
        description="""
        Case insensitive name of the Git provider. Following values are supported right now (could be a subject for 
        change, consult 
        [Repos API documentation](https://docs.databricks.com/dev-tools/api/latest/repos.html)):
        `gitHub`, `gitHubEnterprise`, `bitbucketCloud`, `bitbucketServer`, `azureDevOpsServices`, `gitLab`, 
        `gitLabEnterpriseEdition`.
        """,
    )
    tag: str = Field(
        None,
        description="Name of the Git branch to use. Conflicts with `branch` and `commit`.",
    )


class JobHealthRule(BaseModel):
    metric: str = Field(
        None,
        description="""
    Metric to check. The only supported metric is RUN_DURATION_SECONDS (check Jobs REST API documentation for the latest
    information).
    """,
    )
    op: str = Field(
        None,
        description="""
    Operation used to compare operands. Currently, following operators are supported: EQUAL_TO, GREATER_THAN, 
    GREATER_THAN_OR_EQUAL, LESS_THAN, LESS_THAN_OR_EQUAL, NOT_EQUAL.
    """,
    )
    value: int = Field(None, description="Value used to compare to the given metric.")


class JobHealth(BaseModel):
    rules: list[JobHealthRule] = Field(
        None, description="Job health rules specifications"
    )


class JobNotificationSettings(BaseModel):
    no_alert_for_canceled_runs: bool = Field(
        None, description="If `True`, don't send alert for cancelled runs."
    )
    no_alert_for_skipped_runs: bool = Field(
        None, description="If `True`, don't send alert for skipped runs."
    )


class JobParameter(BaseModel):
    default: str = Field(None, description="Default value of the parameter.")
    name: str = Field(
        None,
        description="The name of the defined parameter. May only contain alphanumeric characters, `_`, `-`, and `.`,",
    )


class JobRunAs(BaseModel):
    service_principal_name: str = Field(
        None,
        description="The application ID of an active service principal. Setting this field requires the servicePrincipal/user role.",
    )
    user_name: str = Field(
        None,
        description="The email of an active workspace user. Non-admin users can only set this field to their own email.",
    )


class JobSchedule(BaseModel):
    quartz_cron_expression: str = Field(
        ...,
        description="A Cron expression using Quartz syntax that describes the schedule for a job. This field is required.",
    )
    timezone_id: str = Field(
        ...,
        description="""
    A Java timezone ID. The schedule for a job will be resolved with respect to this timezone. See Java TimeZone for 
    details. This field is required.
    """,
    )
    pause_status: Union[Literal["PAUSED", "UNPAUSED"], str, None] = Field(
        None,
        description="""
        Indicate whether this schedule is paused or not. When the pause_status field is omitted and a schedule is
        provided, the server will default to using `UNPAUSED` as a value for pause_status.
        """,
    )


class JobTaskConditionTask(BaseModel):
    left: str = Field(
        None,
        description="The left operand of the condition task. It could be a string value, job state, or a parameter reference.",
    )
    op: Literal[
        "EQUAL_TO",
        "GREATER_THAN",
        "GREATER_THAN_OR_EQUAL",
        "LESS_THAN",
        "LESS_THAN_OR_EQUAL",
        "NOT_EQUAL",
    ] = Field(
        None,
        description="""
        The string specifying the operation used to compare operands. This task does not require a cluster to execute
        and does not support retries or notifications.
        """,
    )
    right: str = Field(
        None,
        description="The right operand of the condition task. It could be a string value, job state, or parameter reference.",
    )


class JobTaskDependsOn(BaseModel):
    task_key: str = Field(
        None, description="The name of the task this task depends on."
    )
    outcome: Literal["true", "false"] = Field(
        None,
        description="""
    Can only be specified on condition task dependencies. The outcome of the dependent task that must be met for this
    task to run.""",
    )


class JobTaskNotebookTask(BaseModel):
    notebook_path: str = Field(
        ...,
        description="""
    The path of the databricks.Notebook to be run in the Databricks workspace or remote repository. For notebooks
    stored in the Databricks workspace, the path must be absolute and begin with a slash. For notebooks stored in a
    remote repository, the path must be relative.
    """,
    )
    base_parameters: dict[str, Any] = Field(
        None,
        description="""
    Base parameters to be used for each run of this job. If the run is initiated by a call to run-now with
    parameters specified, the two parameters maps will be merged. If the same key is specified in base_parameters
    and in run-now, the value from run-now will be used. If the notebook takes a parameter that is not specified
    in the job’s base_parameters or the run-now override parameters, the default value from the notebook will be
    used. Retrieve these parameters in a notebook using dbutils.widgets.get.
    """,
    )
    warehouse_id: str = Field(
        None,
        description="""
    The id of the SQL warehouse to execute this task. If a warehouse_id is specified, that SQL warehouse will be
    used to execute SQL commands inside the specified notebook.
    """,
    )
    source: Literal["WORKSPACE", "GIT"] = Field(
        None,
        description="""
    Location type of the notebook, can only be WORKSPACE or GIT. When set to WORKSPACE, the notebook will be
    retrieved from the local Databricks workspace. When set to GIT, the notebook will be retrieved from a Git
    repository defined in git_source. If the value is empty, the task will use GIT if git_source is defined and
    WORKSPACE otherwise.
    """,
    )


class JobTaskPipelineTask(BaseModel):
    pipeline_id: str = Field(None, description="The pipeline's unique ID.")
    full_refresh: bool = Field(
        None, description="Specifies if there should be full refresh of the pipeline."
    )


class JobTaskDbtTask(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    commands: list[str] = Field(
        ...,
        description="Series of dbt commands to execute in sequence. Every command must start with 'dbt'.",
    )
    catalog: str = Field(
        None, description="The name of the catalog to use inside Unity Catalog."
    )
    profiles_directory: str = Field(
        None,
        description="""
    The relative path to the directory in the repository specified by `git_source` where dbt should look in for the 
    `profiles.yml` file. If not specified, defaults to the repository's root directory. Equivalent to passing 
    `--profile-dir` to a dbt command.
    """,
    )
    project_directory: str = Field(
        None,
        description="""
    The path where dbt should look for dbt_project.yml. Equivalent to passing `--project-dir` to the dbt CLI.
    - If source is GIT: Relative path to the directory in the repository specified
      in the git_source block. Defaults to the repository's root directory when not
      specified.
    - If source is WORKSPACE: Absolute path to the folder in the workspace.    
    """,
    )
    schema_: str = Field(
        None,
        alias="schema",
        description="The name of the schema dbt should run in. Defaults to `default`.",
    )  # required not to overwrite BaseModel attribute
    source: str = Field(
        None,
        description="""
    The source of the project. Possible values are `WORKSPACE` and `GIT`. Defaults to `GIT` if a `git_source` block is 
    present in the job definition.
    """,
    )
    warehouse_id: str = Field(
        None, description="The ID of the SQL warehouse that dbt should execute against."
    )


class JobTaskRunJobTask(BaseModel):
    job_id: Union[int, str] = Field(None, description="ID of the job")
    job_parameters: dict[str, Any] = Field(
        None, description="Job parameters for the task"
    )


class JobTaskSqlTaskQuery(BaseModel):
    query_id: str = Field(None, description="Query ID")


class JobTaskSqlTaskAlertSubscription(BaseModel):
    destination_id: str = Field(None, description="")
    user_name: str = Field(
        None,
        description="The email of an active workspace user. Non-admin users can only set this field to their own email.",
    )


class JobTaskSQLTaskAlert(BaseModel):
    alert_id: str = Field(None, description="Identifier of the Databricks SQL Alert.")
    subscriptions: list[JobTaskSqlTaskAlertSubscription] = Field(
        None,
        description="""
    A list of subscription blocks consisting out of one of the required fields: `user_name` for user emails or 
    `destination_id` - for Alert destination's identifier.
    """,
    )
    pause_subscriptions: bool = Field(
        None, description="It `True` subscriptions are paused"
    )


class JobTaskSqlTaskDashboard(BaseModel):
    dashboard_id: str = Field(
        None,
        description="identifier of the Databricks SQL Dashboard databricks_sql_dashboard.",
    )
    custom_subject: list[JobTaskSqlTaskAlertSubscription] = Field(
        None, description="Custom subject specifications"
    )
    subscriptions: list[JobTaskSqlTaskAlertSubscription] = Field(
        None, description="Subscriptions specifications"
    )


class JobTaskSqlTaskFile(BaseModel):
    path: str = Field(
        None,
        description="""
    If source is `GIT`: Relative path to the file in the repository specified in the git_source block with SQL commands 
    to execute. If source is `WORKSPACE`: Absolute path to the file in the workspace with SQL commands to execute.
    """,
    )
    source: Literal["WORKSPACE", "GIT"] = Field(
        None,
        description="The source of the project. Possible values are `WORKSPACE` and `GIT`.",
    )


class JobTaskSQLTask(BaseModel):
    alert: JobTaskSQLTaskAlert = Field(None, description="Alert specifications")
    dashboard: JobTaskSqlTaskDashboard = Field(
        None, description="Dashboard specifications"
    )
    file: JobTaskSqlTaskFile = Field(None, description="File specifications")
    parameters: dict[str, Any] = Field(None, description="Parameters specifications")
    query: JobTaskSqlTaskQuery = Field(None, description="Query specifications")
    warehouse_id: str = Field(..., description="Warehouse id")


class JobTaskPythonWheelTask(BaseModel):
    entry_point: str = Field(
        ..., description="Python function as entry point for the task"
    )
    named_parameters: dict[str, str] = Field(
        None, description="Named parameters for the task"
    )
    package_name: str = Field(None, description="Name of Python package")
    parameters: list[str] = Field(None, description="Parameters for the task")


class JobTaskForEachTaskTask(BaseModel):
    # compute_key: str = Field(None, description="")
    dbt_task: JobTaskDbtTask = Field(None, description="dbt task")
    condition_task: JobTaskConditionTask = Field(
        None, description="Condition Task specifications"
    )
    depends_ons: list[JobTaskDependsOn] = Field(
        None, description="Depends On specifications"
    )
    description: str = Field(None, description="Description for this task")
    email_notifications: JobEmailNotifications = Field(
        None, description="Email Notifications specifications"
    )
    environment_key: str = Field(
        None,
        description="""
    Identifier of an `environment` that is used to specify libraries. Required for some tasks (`spark_python_task`, 
    `python_wheel_task`, …) running on serverless compute.
    """,
    )
    existing_cluster_id: str = Field(
        None,
        description="Cluster id from one of the clusters available in the workspace",
    )
    health: JobHealth = Field(None, description="Job Health specifications")
    job_cluster_key: str = Field(
        None,
        description="Identifier that can be referenced in task block, so that cluster is shared between tasks",
    )
    libraries: list[ClusterLibrary] = Field(
        None, description="Cluster Library specifications"
    )
    max_retries: int = Field(
        None,
        description="An optional maximum number of times to retry an unsuccessful run.",
    )
    min_retry_interval_millis: int = Field(
        None,
        description="""
    An optional minimal interval in milliseconds between the start of the failed run and the subsequent retry run.
    The default behavior is that unsuccessful runs are immediately retried.
    """,
    )
    # new_cluster: Cluster = Field(None, description="")
    notebook_task: JobTaskNotebookTask = Field(
        None, description="Notebook Task specifications"
    )
    notification_settings: JobNotificationSettings = Field(
        None, description="Notification Settings specifications"
    )
    pipeline_task: JobTaskPipelineTask = Field(
        None, description="Pipeline Task specifications"
    )
    python_wheel_task: JobTaskPythonWheelTask = Field(None, description="")
    retry_on_timeout: bool = Field(
        None,
        description="If `True`, retry a job when it times out. The default behavior is to not retry on timeout.",
    )
    run_if: str = Field(
        None,
        description="""
    An optional value indicating the condition that determines whether the task should be run once its dependencies
    have been completed. When omitted, defaults to `ALL_SUCCESS`.
    """,
    )
    run_job_task: JobTaskRunJobTask = Field(None, description="Run Job specifications")
    # spark_jar_task:
    # spark_python_task:
    sql_task: JobTaskSQLTask = Field(None, description="SQL Task specifications")
    task_key: str = Field(None, description="A unique key for a given task.")
    timeout_seconds: int = Field(
        None,
        description="An optional timeout applied to each run of this job. The default behavior is to have no timeout.",
    )

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "email_notifications": "email_notifications",
            "notification_settings": "notification_settings",
            "webhook_notifications": "webhook_notifications",
        }

    @field_validator("depends_ons")
    @classmethod
    def sort_depends_ons(cls, v: list[JobTaskDependsOn]) -> list[JobTaskDependsOn]:
        return sorted(v, key=lambda task: task.task_key)


class JobTaskForEachTask(BaseModel):
    inputs: str | list = Field(
        ...,
        description="""
    Array for task to iterate on. This can be a JSON string or a reference
    to an array parameter. Laktory also supports a list input, which wil
    be serialized.
    """,
    )
    task: JobTaskForEachTaskTask = Field(
        ..., description="Task to run against the inputs list."
    )
    concurrency: int = Field(
        None,
        description="""
    Controls the number of active iteration task runs. Default is 20, maximum allowed is 100.
    """,
    )

    @field_validator("inputs")
    def parse_inputs(cls, v: Union[str, list]) -> str:
        if isinstance(v, list):
            v = json.dumps(v)

        v = v.replace("'", '"')

        return v


class JobTask(JobTaskForEachTaskTask):
    for_each_task: JobTaskForEachTask = Field(
        None, description="For each task configuration"
    )


class JobTriggerFileArrival(BaseModel):
    url: str = Field(None, description="URL of the job on the given workspace")
    min_time_between_triggers_seconds: int = Field(
        None,
        description="""
    If set, the trigger starts a run only after the specified amount of time passed since the last time the trigger 
    fired. The minimum allowed value is 60 seconds.
    """,
    )
    wait_after_last_change_seconds: int = Field(
        None,
        description="""
    If set, the trigger starts a run only after no file activity has occurred for the specified amount of time.
    This makes it possible to wait for a batch of incoming files to arrive before triggering a run.
    The minimum allowed value is 60 seconds.
    """,
    )


class JobTrigger(BaseModel):
    file_arrival: JobTriggerFileArrival = Field(
        ..., description="File Arrival specifications"
    )
    pause_status: Union[Literal["PAUSED", "UNPAUSED"], str] = Field(
        None,
        description="""
        Indicate whether this trigger is paused or not. When the pause_status field is omitted in the block, the server
        will default to using `UNPAUSED` as a value for pause_status.
        """,
    )


class JobWebhookNotificationsOnDurationWarningThresholdExceeded(BaseModel):
    id: str = Field(None, description="Unique identifier")


class JobWebhookNotificationsOnFailure(BaseModel):
    id: str = Field(None, description="Unique identifier")


class JobWebhookNotificationsOnStart(BaseModel):
    id: str = Field(None, description="Unique identifier")


class JobWebhookNotificationsOnSuccess(BaseModel):
    id: str = Field(None, description="Unique identifier")


class JobWebhookNotifications(BaseModel):
    on_duration_warning_threshold_exceededs: list[
        JobWebhookNotificationsOnDurationWarningThresholdExceeded
    ] = Field(None, description="Warnings threshold exceeded specifications")
    on_failures: list[JobWebhookNotificationsOnFailure] = Field(
        None, description="On failure specifications"
    )
    on_starts: list[JobWebhookNotificationsOnStart] = Field(
        None, description="On starts specifications"
    )
    on_successes: list[JobWebhookNotificationsOnSuccess] = Field(
        None, description="On successes specifications"
    )


class JobQueue(BaseModel):
    enabled: bool = Field(..., description="If true, enable queueing for the job.")


class JobLookup(ResourceLookup):
    id: str = Field(
        serialization_alias="id", description="The id of the databricks job"
    )


class Job(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Job

    Examples
    --------
    ```py
    import io

    from laktory import models

    # Define job
    job_yaml = '''
    name: job-stock-prices
    clusters:
      - name: main
        spark_version: 16.3.x-scala2.12
        node_type_id: Standard_DS3_v2

    tasks:
      - task_key: ingest
        job_cluster_key: main
        notebook_task:
          notebook_path: /jobs/ingest_stock_prices.py
        libraries:
          - pypi:
              package: yfinance

      - task_key: pipeline
        depends_ons:
          - task_key: ingest
        pipeline_task:
          pipeline_id: 74900655-3641-49f1-8323-b8507f0e3e3b

    access_controls:
      - group_name: account users
        permission_level: CAN_VIEW
      - group_name: role-engineers
        permission_level: CAN_MANAGE_RUN
    '''
    job = models.resources.databricks.Job.model_validate_yaml(io.StringIO(job_yaml))

    # Define job with for each task
    job_yaml = '''
    name: job-hello
    tasks:
      - task_key: hello-loop
        for_each_task:
          inputs:
            - id: 1
              name: olivier
            - id: 2
              name: kubic
          task:
            task_key: hello-task
            notebook_task:
              notebook_path: /Workspace/Users/olivier.soucy@okube.ai/hello-world
              base_parameters:
                input: "{{input}}"
    '''
    job = models.resources.databricks.Job.model_validate_yaml(io.StringIO(job_yaml))
    ```

    References
    ----------

    * [Databricks Job](https://docs.databricks.com/en/workflows/jobs/create-run-jobs.html)
    * [Pulumi Databricks Job](https://www.pulumi.com/registry/packages/databricks/api-docs/job/#databricks-job)
    """

    access_controls: list[AccessControl] = Field([], description="Access controls list")
    continuous: JobContinuous = Field(None, description="Continuous specifications")
    control_run_state: bool = Field(
        None,
        description="""
    If `True`, the Databricks provider will stop and start the job as needed to ensure that the active run for the
    job reflects the deployed configuration. For continuous jobs, the provider respects the pause_status by
    stopping the current active run. This flag cannot be set for non-continuous jobs.
    """,
    )
    description: str = Field(
        None,
        description="An optional description for the job. The maximum length is 1024 characters in UTF-8 encoding.",
    )
    email_notifications: JobEmailNotifications = Field(
        None,
        description="""
    An optional set of email addresses notified when runs of this job begins, completes or fails. The default
    behavior is to not send any emails. This field is a block and is documented below.
    """,
    )
    environments: list[JobEnvironment] = Field(
        None, description="List of environments available for the tasks."
    )
    format: str = Field(None, description="")
    git_source: JobGitSource = Field(
        None, description="Specifies a Git repository for task source code."
    )
    health: JobHealth = Field(None, description="Health specifications")
    job_clusters: list[JobJobCluster] = Field(
        [],
        description="""
    A list of job databricks.Cluster specifications that can be shared and reused by tasks of this job. Libraries 
    cannot be declared in a shared job cluster. You must declare dependent libraries in task settings.
    """,
    )

    lookup_existing: JobLookup = Field(
        None,
        exclude=True,
        description="Specifications for looking up existing resource. Other attributes will be ignored.",
    )
    max_concurrent_runs: int = Field(
        None,
        description="An optional maximum allowed number of concurrent runs of the job. Defaults to 1.",
    )
    max_retries: int = Field(
        None,
        description="""
    An optional maximum number of times to retry an unsuccessful run. A run is considered to be unsuccessful if it
    completes with a FAILED or INTERNAL_ERROR lifecycle state. The value -1 means to retry indefinitely and the
    value 0 means to never retry. The default behavior is to never retry. A run can have the following lifecycle
    state: PENDING, RUNNING, TERMINATING, TERMINATED, SKIPPED or INTERNAL_ERROR.
    """,
    )
    min_retry_interval_millis: int = Field(
        None,
        description="""
    An optional minimal interval in milliseconds between the start of the failed run and the subsequent retry run.
    The default behavior is that unsuccessful runs are immediately retried.
    """,
    )
    name: str = Field(None, description="Name of the job")
    name_prefix: str = Field(None, description="Prefix added to the job name")
    name_suffix: str = Field(None, description="Suffix added to the job name")
    notification_settings: JobNotificationSettings = Field(
        None, description="Notifications specifications"
    )
    parameters: list[JobParameter] = Field([], description="Parameters specifications")
    queue: JobQueue = Field(None, description="")
    retry_on_timeout: bool = Field(
        None,
        description="""
        An optional policy to specify whether to retry a job when it times out. The default behavior is to not retry on
        timeout.
        """,
    )
    run_as: JobRunAs = Field(None, description="Run as specifications")
    schedule: JobSchedule = Field(None, description="Schedule specifications")
    tags: dict[str, Any] = Field({}, description="Tags as key, value pairs")
    tasks: list[JobTask] = Field([], description="Tasks specifications")
    timeout_seconds: int = Field(
        None,
        description="An optional timeout applied to each run of this job. The default behavior is to have no timeout.",
    )
    trigger: JobTrigger = Field(None, description="Trigger specifications")
    webhook_notifications: JobWebhookNotifications = Field(
        None, description="Webhook notifications specifications"
    )

    @field_validator("tasks")
    @classmethod
    def sort_tasks(cls, v: list[JobTask]) -> list[JobTask]:
        return sorted(v, key=lambda task: task.task_key)

    @model_validator(mode="after")
    def update_name(self) -> Any:
        with self.validate_assignment_disabled():
            if self.name_prefix:
                self.name = self.name_prefix + self.name
                self.name_prefix = ""
            if self.name_suffix:
                self.name = self.name + self.name_suffix
                self.name_suffix = ""

        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    job_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Job"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls", "name_prefix", "name_suffix"]

    @property
    def pulumi_properties(self):
        d = super().pulumi_properties

        # Rename dbt task schema
        for task in d["tasks"]:
            if "dbt_task" in task:
                if "schema_" in task["dbt_task"]:
                    task["dbt_task"]["schema"] = task["dbt_task"]["schema_"]
                    del task["dbt_task"]["schema_"]

        return d

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "email_notifications": "email_notifications",
            "notification_settings": "notification_settings",
            "webhook_notifications": "webhook_notifications",
        }

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_job"

    @property
    def terraform_excludes(self) -> list[str] | dict[str, bool]:
        return self.pulumi_excludes

    @property
    def terraform_properties(self) -> dict:
        d = super().terraform_properties

        # Rename dbt task schema
        if "tasks" in d:
            for task in d["tasks"]:
                if "dbt_task" in task:
                    if "schema_" in task["dbt_task"]:
                        task["dbt_task"]["schema"] = task["dbt_task"]["schema_"]
                        del task["dbt_task"]["schema_"]

        return d



from functools import wraps

import narwhals as nw

from laktory.narwhals_ext.functions.current_timestamp import current_timestamp
from laktory.narwhals_ext.functions.sql_expr import sql_expr


class LaktoryFuncs:  # noqa: F811
    @wraps(current_timestamp)
    def current_timestamp(*args, **kwargs):
        return current_timestamp(*args, **kwargs)

    @wraps(sql_expr)
    def sql_expr(*args, **kwargs):
        return sql_expr(*args, **kwargs)


nw.laktory = LaktoryFuncs



from typing import Callable

from pydantic import Field
from pydantic import computed_field

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource

logger = get_logger(__name__)


class PythonPackagePermissions(BaseModel, PulumiResource, TerraformResource):
    access_controls: list[AccessControl] = Field(
        ..., description="Access controls list"
    )
    get_workspace_file_path: Callable = Field(
        ..., description="Callable returning workspace filepath", exclude=True
    )

    @computed_field(description="workspace_file_path")
    @property
    def workspace_file_path(self) -> str:
        return self.get_workspace_file_path()

    # ----------------------------------------------------------------------- #
    # Pulumi Methods                                                          #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Permissions"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_permissions"



# import abc
from typing import Any

# from typing import Callable
# from typing import Literal
from typing import Union

import narwhals as nw
from pydantic import Field
from pydantic import model_validator

from laktory._logger import get_logger
from laktory.enums import DataFrameBackends
from laktory.models.basemodel import BaseModel
from laktory.models.datasources import DataSourcesUnion
from laktory.models.pipelinechild import PipelineChild
from laktory.typing import AnyFrame

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Helper Classes                                                              #
# --------------------------------------------------------------------------- #


class DataFrameMethodArg(BaseModel, PipelineChild):
    """
    DataFrame method argument expressed as a string or a serialized DataSource.
    """

    value: DataSourcesUnion | Any = Field(..., description="Function argument")

    def eval(self, backend: DataFrameBackends):
        from laktory.models.datasources.basedatasource import BaseDataSource

        # TODO: Add supported for evaluating list or dict of strings.
        v = self.value

        if isinstance(v, BaseDataSource):
            v = self.value.read()
            if self.dataframe_api == "NATIVE":
                v = v.to_native()

        elif isinstance(v, str):
            # Imports required to evaluate expressions
            if self.dataframe_api == "NARWHALS":
                import narwhals as nw  # noqa: F401
                from narwhals import col  # noqa: F401
                from narwhals import lit  # noqa: F401

                from laktory.narwhals_ext.functions import sql_expr  # noqa: F401

                targets = ["lit(", "col(", "nw.", "sql_expr"]

                # TODO: Review if we want to ducktype narwhals
                v = v.replace("nw.sql_expr", "sql_expr")

            else:
                from laktory.enums import DataFrameBackends

                if backend == DataFrameBackends.PYSPARK:
                    # Imports required to evaluate expressions
                    import pyspark.sql.functions as F  # noqa: F401
                    from pyspark.sql.functions import col  # noqa: F401
                    from pyspark.sql.functions import expr  # noqa: F401
                    from pyspark.sql.functions import lit  # noqa: F401

                    targets = ["lit(", "col(", "expr(", "F."]

                elif backend == DataFrameBackends.POLARS:
                    # Imports required to evaluate expressions
                    import polars as pl  # noqa: F401
                    from polars import col  # noqa: F401
                    from polars import lit  # noqa: F401
                    from polars import sql_expr  # noqa: F401

                    targets = ["lit(", "col(", "sql_expr(", "pl."]

                else:
                    raise NotImplementedError()

            for f in targets:
                if f in v:
                    v = eval(v)
                    break

        return v

    def signature(self):
        from laktory.models.datasources import DataFrameDataSource
        from laktory.models.datasources import FileDataSource
        from laktory.models.datasources import PipelineNodeDataSource
        from laktory.models.datasources import TableDataSource

        if isinstance(self.value, DataFrameDataSource):
            return f"{self.value.df}"
        elif isinstance(self.value, PipelineNodeDataSource):
            return f"node.{self.value.node_name}"
        elif isinstance(self.value, FileDataSource):
            return f"file {self.value.path}"
        elif isinstance(self.value, TableDataSource):
            return f"table {self.value.full_name}"
        return str(self.value)


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class DataFrameMethod(BaseModel, PipelineChild):
    """
    Definition of a DataFrame method to be applied. Both native and Narwhals API are
    supported.

    Examples
    --------
    ```py
    import polars as pl

    import laktory as lk

    df0 = pl.DataFrame(
        {
            "x": [1.1, 2.2, 3.3],
        }
    )

    m1 = lk.models.DataFrameMethod(
        func_name="with_columns",
        func_kwargs={"xr": "nw.col('x').round()"},
        dataframe_api="NARWHALS",
    )
    df = m1.execute(df0)

    m2 = lk.models.DataFrameMethod(
        func_name="select", func_args=["pl.col('x').sqrt()"], dataframe_api="NATIVE"
    )
    df = m2.execute(df0)

    print(df.to_native())
    '''
    | x        |
    |----------|
    | 1.048809 |
    | 1.48324  |
    | 1.81659  |
    '''
    ```
    """

    func_args: list[DataFrameMethodArg | Any] = Field(
        [],
        description="Arguments passed to method. A `DataSource` model can be passed instead of a DataFrame.",
    )
    func_kwargs: dict[str, DataFrameMethodArg | Any] = Field(
        {},
        description="Keyword arguments passed to method. A `DataSource` model can be passed instead of a DataFrame.",
    )
    func_name: str = Field(..., description="Method name.")

    @model_validator(mode="after")
    def set_args(self) -> Any:
        for k, v in self.func_kwargs.items():
            if not isinstance(v, DataFrameMethodArg):
                self.func_kwargs[k] = DataFrameMethodArg(value=v)
                # Because we don't set func_kwargs directly,
                # assign_parent_to_children() is not triggered and we need to set
                # parent explicitly
                self.func_kwargs[k].parent = self

        for i, v in enumerate(self.func_args):
            if not isinstance(v, DataFrameMethodArg):
                self.func_args[i] = DataFrameMethodArg(value=v)
                # Because we don't set func_kwargs directly,
                # assign_parent_to_children() is not triggered and we need to set
                # parent explicitly
                self.func_args[i].parent = self

        return self

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def children_names(self):
        return [
            # "data_sources",
            "func_args",
            "func_kwargs",
            # "sql_expr",
        ]

    #
    # # ----------------------------------------------------------------------- #
    # # Columns Creation                                                        #
    # # ----------------------------------------------------------------------- #
    #
    # @property
    # def _with_columns(self) -> list[ChainNodeColumn]:
    #     with_columns = [c for c in self.with_columns]
    #     if self.with_column:
    #         with_columns += [self.with_column]
    #     return with_columns
    #
    # @property
    # def is_column(self):
    #     return len(self._with_columns) > 0
    #
    # ----------------------------------------------------------------------- #
    # Data Sources                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def data_sources(self) -> list[DataSourcesUnion]:
        """Get all sources feeding the DataFrame Method"""

        from laktory.models.datasources import BaseDataSource

        sources = []
        for a in self.func_args:
            if isinstance(a.value, BaseDataSource):
                sources += [a.value]

        for a in self.func_kwargs.values():
            if isinstance(a.value, BaseDataSource):
                sources += [a.value]

        return sources

    # ----------------------------------------------------------------------- #
    # Upstream Nodes                                                          #
    # ----------------------------------------------------------------------- #

    @property
    def upstream_node_names(self) -> list[str]:
        """Pipeline node names required to apply transformer node."""

        from laktory.models.datasources.pipelinenodedatasource import (
            PipelineNodeDataSource,
        )

        names = []
        for a in self.func_args:
            if isinstance(a.value, PipelineNodeDataSource):
                names += [a.value.node_name]
        for a in self.func_kwargs.values():
            if isinstance(a.value, PipelineNodeDataSource):
                names += [a.value.node_name]

        return names

    # ----------------------------------------------------------------------- #
    # Execution                                                               #
    # ----------------------------------------------------------------------- #

    def execute(self, df: AnyFrame) -> Union[AnyFrame]:
        """
        Execute method on provided DataFrame `df`.

        Parameters
        ----------
        df:
            Input dataframe

        Returns
        -------
            Output dataframe
        """

        # Get and set Backend (required to evaluate arguments)
        backend = DataFrameBackends.from_df(df)
        self.dataframe_backend_ = backend

        # Convert to Narwhals
        if not isinstance(df, AnyFrame):
            df = nw.from_native(df)
        if self.dataframe_api == "NATIVE":
            df = df.to_native()

        # Get Function
        namespace = None
        func_name = self.func_name
        func_full_name = func_name
        if "." in func_name:
            namespace, func_name = func_name.split(".")
        df_as_input = False

        # Get from built-in narwhals and narwhals extension (including Laktory) functions
        f = None
        if f is None:
            # Get function from namespace extension
            if namespace:
                f = getattr(getattr(df, namespace), func_name, None)
            else:
                f = getattr(df, func_name, None)

        if f is None:
            df_type = type(df)
            raise ValueError(
                f"Function {func_full_name} is not available on dataframe of type {str(df_type)} with {self.dataframe_api} API"
            )

        _args = self.func_args
        _kwargs = self.func_kwargs

        # Build log
        func_log = f"df.{func_full_name}("
        func_log += ",".join([a.signature() for a in _args])
        func_log += ",".join([f"{k}={a.signature()}" for k, a in _kwargs.items()])
        func_log += f") with df type {type(df)}"
        logger.info(f"Applying {func_log}")

        # Build args
        args = []
        if df_as_input:
            args += [df]
        for i, _arg in enumerate(_args):
            args += [_arg.eval(backend=backend)]

        # Build kwargs
        kwargs = {}
        for k, _arg in _kwargs.items():
            kwargs[k] = _arg.eval(backend=backend)

        # Call function
        df = f(*args, **kwargs)

        # Convert to narwhals when custom function don't return a Narwhals DataFrame
        if not isinstance(df, AnyFrame):
            df = nw.from_native(df)

        return df

    # def get_view_definition(self):
    #     return self._parsed_sql_expr.parsed_expr(view=True)


# TransformerFuncArg.model_rebuild()



from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.mwsnccbinding import MwsNccBinding
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class MwsNetworkConnectivityConfigEgressConfigDefaultRulesAwsStableIpRule(BaseModel):
    cidr_blocks: list[str] = Field(None, description="")


class MwsNetworkConnectivityConfigEgressConfigDefaultRulesAzureServiceEndpointRule(
    BaseModel
):
    subnets: list[str] = Field(None, description="")
    target_region: str = Field(None, description="")
    target_services: list[str] = Field(None, description="")


class MwsNetworkConnectivityConfigEgressConfigDefaultRules(BaseModel):
    aws_stable_ip_rule: MwsNetworkConnectivityConfigEgressConfigDefaultRulesAwsStableIpRule = Field(
        None, description=""
    )
    azure_service_endpoint_rule: MwsNetworkConnectivityConfigEgressConfigDefaultRulesAzureServiceEndpointRule = Field(
        None,
        description="""
    This provides a list of subnets. These subnets need to be allowed in your Azure resources in order for 
    Databricks to access.
    """,
    )


class MwsNetworkConnectivityConfigEgressConfigTargetRulesAzurePrivateEndpointRule(
    BaseModel
):
    connection_state: str = Field(None, description="")
    creation_time: int = Field(None, description="")
    deactivated: bool = Field(None, description="")
    deactivated_at: int = Field(None, description="")
    endpoint_name: str = Field(None, description="")
    group_id: str = Field(None, description="")
    network_connectivity_config_id: str = Field(
        None,
        description="Canonical unique identifier of Network Connectivity Config in Databricks Account",
    )
    resource_id: str = Field(None, description="")
    rule_id: str = Field(None, description="")
    updated_time: int = Field(None, description="")


class MwsNetworkConnectivityConfigEgressConfigTargetRules(BaseModel):
    azure_private_endpoint_rules: list[
        MwsNetworkConnectivityConfigEgressConfigTargetRulesAzurePrivateEndpointRule
    ] = Field(None, description="")


class MwsNetworkConnectivityConfigEgressConfig(BaseModel):
    default_rules: MwsNetworkConnectivityConfigEgressConfigDefaultRules = Field(
        None, description=""
    )
    target_rules: MwsNetworkConnectivityConfigEgressConfigTargetRules = Field(
        None, description=""
    )


class MwsNetworkConnectivityConfig(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Mws Network Connectivity Config

    Examples
    --------
    ```py
    ```
    """

    account_id: str = Field(None, description="")
    creation_time: int = Field(None, description="")
    egress_config: MwsNetworkConnectivityConfigEgressConfig = Field(
        None, description="Egress Config configration"
    )
    name: str = Field(
        ...,
        description="Name of Network Connectivity Config in Databricks Account. Change forces creation of a new resource.",
    )
    network_connectivity_config_id: str = Field(
        None,
        description="Canonical unique identifier of Network Connectivity Config in Databricks Account",
    )
    region: str = Field(
        ...,
        description="""
    Region of the Network Connectivity Config. NCCs can only be referenced by your workspaces in the same region.
    Change forces creation of a new resource.
    """,
    )
    updated_time: int = Field(None, description="")
    workspace_bindings: list[MwsNccBinding] = Field(None, description="")

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - workspace bindings
        """
        resources = []

        if self.workspace_bindings:
            for b in self.workspace_bindings:
                b.network_connectivity_config_id = f"${{resources.{self.resource_name}.network_connectivity_config_id}}"
                resources += [b]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:MwsNetworkConnectivityConfig"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return [
            "workspace_bindings",
        ]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_mws_network_connectivity_config"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import narwhals as nw


def schema_flat(self) -> list[str]:
    """
    Returns a flattened list of columns

    Returns
    -------
    :
        List of columns

    Examples
    --------
    ```py
    import narwhals as nw
    import polars as pl

    import laktory as lk  # noqa: F401

    df = nw.from_native(
        pl.DataFrame(
            {
                "index": [1, 2, 3],
                "stock": [
                    {"symbol": "AAPL", "name": "Apple"},
                    {"symbol": "MSFT", "name": "Microsoft"},
                    {"symbol": "GOOGL", "name": "Google"},
                ],
                "prices": [
                    [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
                    [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
                    [{"open": 1, "close": 2}, {"open": 1, "close": 2}],
                ],
            }
        )
    )

    print(df.laktory.schema_flat())
    '''
    [
        'index',
        'stock',
        'stock.symbol',
        'stock.name',
        'prices',
        'prices[*].open',
        'prices[*].close',
    ]
    '''
    ```
    """

    def get_fields(schema):
        field_names = []
        for f_name, f_type in schema.items():
            if isinstance(f_type, nw.Struct):
                _field_names = get_fields(dict(f_type))
                field_names += [f_name]
                field_names += [f"{f_name}.{v}" for v in _field_names]

            elif isinstance(f_type, nw.List):
                field_names += [f_name]
                if isinstance(f_type.inner, nw.Struct):
                    _field_names = get_fields(dict(f_type.inner))
                    field_names += [f"{f_name}[*].{v}" for v in _field_names]

            else:
                field_names += [f_name]

        return field_names

    return get_fields(self._df.schema)



import os
from pathlib import Path
from typing import Union

from pydantic import AliasChoices
from pydantic import Field
from pydantic import computed_field

from laktory import settings
from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class WorkspaceFile(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Workspace File

    Examples
    --------
    ```py
    from laktory import models

    file = models.resources.databricks.WorkspaceFile(
        source="./notebooks/dlt/dlt_laktory_pl.py",
    )
    print(file.path)
    # > /.laktory/dlt_laktory_pl.py

    file = models.resources.databricks.WorkspaceFile(
        source="./notebooks/dlt/dlt_laktory_pl.py",
    )
    print(file.path)
    # > /.laktory/dlt_laktory_pl.py

    file = models.resources.databricks.WorkspaceFile(
        source="./notebooks/dlt/dlt_laktory_pl.py",
        dirpath="notebooks/dlt/",
    )
    print(file.path)
    # > /.laktory/notebooks/dlt/dlt_laktory_pl.py
    ```
    """

    access_controls: list[AccessControl] = Field([], description="Access controls list")
    dirpath: str = Field(
        None,
        description="Workspace directory inside rootpath in which the workspace file is deployed. Used only if `path` is not specified.",
    )
    path_: str = Field(
        None,
        description="Workspace filepath for the file. Overwrite `dirpath`.",
        validation_alias=AliasChoices("path", "path_"),
        exclude=True,
    )
    source_: str = Field(
        None,
        description="Path to file on local filesystem.",
        validation_alias=AliasChoices("source", "source_"),
        exclude=True,
    )
    content_base64: str = Field(
        None,
        description="""
    The base64-encoded file content. Conflicts with source. Use of content_base64 is discouraged, as it's increasing 
    memory footprint of Pulumi state and should only be used in exceptional circumstances, like creating a workspace 
    file with configuration properties for a data pipeline.
    """,
    )

    @computed_field(description="source")
    @property
    def source(self) -> str:
        return self.source_

    @computed_field(description="path")
    @property
    def path(self) -> str | None:
        if self.path_:
            return self.path_

        if not self.source_:
            return None

        # dir
        if self.dirpath is None:
            self.dirpath = ""
        if self.dirpath.startswith("/"):
            self.dirpath = self.dirpath[1:]

        path = Path(settings.workspace_laktory_root) / self.dirpath / self.filename
        return path.as_posix()

    @classmethod
    def lookup_defaults(cls) -> dict:
        return {"path": ""}

    @property
    def filename(self) -> str | None:
        """File filename"""
        if self.source:
            return os.path.basename(self.source)

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.path

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    # workspace_file_path=f"${{resources.{self.resource_name}.path}}",
                    workspace_file_path=self.path,
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:WorkspaceFile"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls", "dirpath"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_workspace_file"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from pydantic import Field

from laktory.models.resources.providers.baseprovider import BaseProvider
from laktory.models.resources.pulumiresource import PulumiResource


class AzurePulumiProvider(BaseProvider, PulumiResource):
    """
    Azure Pulumi (Native) Provider

    Examples
    --------
    ```py
    from laktory import models

    p = models.AzureProvider(
        client_id="${vars.AZURE_CLIENT_ID}",
        client_secret="${vars.AZURE_CLIENT_SECRET}",
    )
    ```
    """

    auxiliary_tenant_ids: list[str] = Field(None, description="")
    client_certificate_password: str = Field(
        None,
        description="""
    The password associated with the Client Certificate. For use when
    authenticating as a Service Principal using a Client Certificate
    """,
    )
    client_certificate_path: str = Field(
        None,
        description="""
    The path to the Client Certificate associated with the Service Principal for use
    when authenticating as a Service Principal using a Client Certificate.
    """,
    )
    client_id: str = Field(None, description="The Client ID which should be used.")
    client_secret: str = Field(
        None,
        description="""
    The Client Secret which should be used. For use When authenticating as a Service Principal using a Client Secret.
    """,
    )
    disable_pulumi_partner_id: bool = Field(None, description="")
    environment: str = Field(
        None,
        description="""
    The Cloud Environment which should be used. Possible values are public, usgovernment, and china. Defaults to 
    public. It can also be sourced from the following environment variables: AZURE_ENVIRONMENT, ARM_ENVIRONMENT
    """,
    )
    location: str = Field(None, description="")
    metadata_host: str = Field(
        None,
        description="""
    The Hostname which should be used for the Azure Metadata Service. It can also be sourced from the following 
    environment variable: ARM_METADATA_HOSTNAME
    """,
    )
    msi_endpoint: str = Field(
        None,
        description="""
    The path to a custom endpoint for Managed Service Identity - in most circumstances this should be detected 
    automatically.
    """,
    )
    oidc_request_token: str = Field(
        None,
        description="""
    The bearer token for the request to the OIDC provider. For use when authenticating as a Service Principal using 
    OpenID Connect.
    """,
    )
    oidc_request_url: str = Field(
        None,
        description="""
    The URL for the OIDC provider from which to request an ID token. For use when authenticating as a Service Principal 
    using OpenID Connect.
    """,
    )
    oidc_token: str = Field(
        None,
        description="""
    The OIDC ID token for use when authenticating as a Service Principal using OpenID Connect.
    """,
    )
    partner_id: str = Field(
        None,
        description=""""
    A GUID/UUID that is registered with Microsoft to facilitate partner resource usage attribution.
    """,
    )
    subscription_id: str = Field(
        None,
        description="""
    The Subscription ID which should be used. It can also be sourced from the following environment variable: 
    ARM_SUBSCRIPTION_ID
    """,
    )
    tenant_id: str = Field(None, description="The Tenant ID which should be used.")
    use_msi: bool = Field(
        None,
        description="Allow Managed Service Identity to be used for Authentication.",
    )
    use_oidc: bool = Field(
        None, description="Allow OpenID Connect to be used for authentication"
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    # @property
    # def resource_key(self) -> str:
    #     return self.display_name

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "pulumi:providers:azure_native"



import json
from pathlib import Path

from pydantic import computed_field

from laktory._logger import get_logger
from laktory._settings import settings
from laktory.models.pipelinechild import PipelineChild
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.workspacefile import WorkspaceFile

logger = get_logger(__name__)


class PipelineConfigWorkspaceFile(WorkspaceFile, PipelineChild):
    """
    Workspace File storing pipeline configuration. Default values for path and
    access controls. Forced value for source.
    """

    access_controls: list[AccessControl] = [
        AccessControl(permission_level="CAN_READ", group_name="users")
    ]

    @computed_field(description="source")
    @property
    def source(self) -> str | None:
        from laktory._cache import cache_dir

        pl_name = ""
        try:
            pl = self.parent_pipeline
            pl_name = pl.name
        except ImportError:
            # parent pipeline can't be access at initial import
            pass

        source_path = cache_dir / "pipelines" / (pl_name + ".json")

        return str(source_path)

    @computed_field(description="source")
    @property
    def path(self) -> str | None:
        if self.path_:
            return self.path_

        pl = self.parent_pipeline
        if not pl:
            return None

        return f"{settings.workspace_laktory_root}pipelines/{pl.name}.json"

    @property
    def content_dict(self):
        pl = self.parent_pipeline
        if not pl:
            return None

        # Overwrite serialization options
        ss0 = self._singular_serialization
        cs0 = self._camel_serialization
        pl._configure_serializer(singular=False, camel=False)

        # Orchestrator (which includes WorkspaceFile) needs to be excluded to avoid
        # infinite re-cursive loop
        _config = self.inject_vars_into_dump(
            {
                "config": pl.model_dump(
                    exclude_unset=True, exclude="orchestrator", mode="json"
                )
            }
        )["config"]
        _config["orchestrator"] = pl.orchestrator.model_dump(
            exclude_unset=True, exclude="config_file", mode="json"
        )

        # Reset serialization options
        pl._configure_serializer(singular=ss0, camel=cs0)

        return _config

    def build(self):
        """
        Write config file to cache (required for deployment).
        """
        filepath = Path(self.source)
        filepath.parent.mkdir(parents=True, exist_ok=True)
        logger.debug(f"Writing config file at {filepath}")
        with filepath.open(mode="w") as fp:
            json.dump(self.content_dict, fp, indent=4)

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_excludes(self) -> list[str] | dict[str, bool]:
        return super().pulumi_excludes + [
            "dataframe_backend",
            "dataframe_api",
        ]

    @property
    def resource_type_id(self):
        return "workspace-file"



from laktory._logger import get_logger

logger = get_logger(__name__)


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


def _execute():
    """Execute pipeline as a script"""
    # TODO: Refactor and integrate into dispatcher / executor / CLI

    import argparse

    import laktory as lk

    # Parse arguments
    parser = argparse.ArgumentParser(
        description="Read pipeline configuration file and execute"
    )
    parser.add_argument(
        "--filepath", type=str, help="Pipeline configuration filepath", required=True
    )
    parser.add_argument(
        "--node_name",
        type=str,
        help="Node name",
        default=None,
        required=False,
    )
    parser.add_argument(
        "--full_refresh",
        type=str2bool,
        help="Full refresh",
        default=False,
        required=False,
    )

    # Get arguments
    args, unknown = parser.parse_known_args()
    filepath = args.filepath
    node_name = args.node_name
    full_refresh = args.full_refresh
    node_str = ""
    if node_name:
        node_str = f" node '{node_name}' of "
    logger.info(
        f"Executing{node_str} pipeline '{filepath}' with full refresh {full_refresh}"
    )

    # Read
    with open(filepath, "r") as fp:
        if str(filepath).endswith(".yaml"):
            pl = lk.models.Pipeline.model_validate_yaml(fp)
        else:
            pl = lk.models.Pipeline.model_validate_json(fp.read())

    # Execute
    if node_name:
        pl.nodes_dict[node_name].execute(full_refresh=full_refresh)
    else:
        pl.execute(full_refresh=full_refresh)



# import abc
import re
from typing import TYPE_CHECKING
from typing import Literal

# from typing import Callable
# from typing import Literal
import narwhals as nw
from pydantic import Field

from laktory._logger import get_logger
from laktory.enums import DataFrameBackends
from laktory.models.basemodel import BaseModel
from laktory.models.pipelinechild import PipelineChild
from laktory.typing import AnyFrame

logger = get_logger(__name__)

if TYPE_CHECKING:
    from laktory.models.datasources.pipelinenodedatasource import PipelineNodeDataSource


# --------------------------------------------------------------------------- #
# Helper Functions                                                            #
# --------------------------------------------------------------------------- #


def to_safe_expr(expr, df_names=None):
    if df_names is None:
        df_names = []
    df_names += ["df"]

    # Update nodes
    pattern = r"\{nodes\.(.*?)\}"
    repl = r"__nodes_\1___"
    expr = re.sub(pattern, repl, expr)

    # Replace df names
    for df_name in df_names:
        expr = expr.replace("{" + df_name + "}", f"__{df_name}__")

    return expr


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class DataFrameExpr(BaseModel, PipelineChild):
    """
    A DataFrame expressed as a SQL statement.

    Examples
    --------
    ```py
    import polars as pl

    import laktory as lk

    df0 = pl.DataFrame(
        {
            "x": [1, 2, 3],
        }
    )

    expr = lk.models.DataFrameExpr(expr="SELECT x, 2*x AS y FROM {df}")
    df = expr.to_df(dfs={"df": df0}).collect()

    print(df)
    '''
    ┌──────────────────┐
    |Narwhals DataFrame|
    |------------------|
    |    | x | y |     |
    |    |---|---|     |
    |    | 1 | 2 |     |
    |    | 2 | 4 |     |
    |    | 3 | 6 |     |
    └──────────────────┘
    '''
    ```
    """

    expr: str = Field(..., description="SQL Expression")
    type: Literal["SQL"] = Field(
        "SQL",
        description="Expression type. Only SQL is currently supported, but `DF` could be added in the future.",
    )
    _data_sources: list["PipelineNodeDataSource"] = None

    @property
    def upstream_node_names(self) -> list[str]:
        """Get all upstream nodes"""
        if self.expr is None:
            return []

        names = []

        pattern = r"\{nodes\.(.*?)\}"
        matches = re.findall(pattern, self.expr)
        for m in matches:
            names += [m]

        return names

    @property
    def data_sources(self):
        """Get all sources required by SQL Expression"""
        if self._data_sources is None:
            from laktory.models.datasources.pipelinenodedatasource import (
                PipelineNodeDataSource,
            )

            sources = []

            pattern = r"\{nodes\.(.*?)\}"
            matches = re.findall(pattern, self.expr)
            for m in matches:
                sources += [PipelineNodeDataSource(node_name=m)]

            self._data_sources = sources

        return self._data_sources

    def to_sql(self, references=None):
        from laktory.models.datasources.pipelinenodedatasource import (
            PipelineNodeDataSource,
        )
        from laktory.models.datasources.tabledatasource import TableDataSource

        if references is None:
            references = {}

        # Resolve Data Sources (references to other nodes)
        for s in self.data_sources:
            references["{nodes." + s.node.name + "}"] = s.sink_table_full_name

        # Add node data source
        pl_node = self.parent_pipeline_node
        if pl_node and pl_node.source:
            s = pl_node.source
            if isinstance(s, TableDataSource):
                full_name = s.full_name
            elif isinstance(s, PipelineNodeDataSource):
                full_name = s.sink_table_full_name
            else:
                raise ValueError(
                    "VIEW sink only supports Table or Pipeline Node with Table sink data sources"
                )
            references["{df}"] = full_name

        expr = self.expr
        for k, v in references.items():
            expr = expr.replace(k, v)

        return expr
        #
        # from laktory.models.datasources.tabledatasource import TableDataSource
        #
        # expr = self.sql_expr
        #
        # pl_node = self.parent_pipeline_node
        #
        # if pl_node and pl_node.source:
        #     source = pl_node.source
        #     if isinstance(source, TableDataSource):
        #         full_name = source.full_name
        #     elif isinstance(source, PipelineNodeDataSource):
        #         full_name = source.sink_table_full_name
        #     else:
        #         raise ValueError(
        #             "VIEW sink only supports Table or Pipeline Node with Table sink data sources"
        #         )
        #     expr = expr.replace("{df}", full_name)
        #
        # pl = self.parent_pipeline
        # if pl:
        #     from laktory.models.datasinks.tabledatasink import TableDataSink
        #
        #     pattern = r"\{nodes\.(.*?)\}"
        #     matches = re.findall(pattern, expr)
        #     for m in matches:
        #         if m not in pl.nodes_dict:
        #             raise ValueError(
        #                 f"Node '{m}' is not available from pipeline '{pl.name}'"
        #             )
        #         sink = pl.nodes_dict[m].primary_sink
        #         if not isinstance(sink, TableDataSink):
        #             raise ValueError(
        #                 f"Node '{m}' used in view creation does not have a Table sink"
        #             )
        #         expr = expr.replace("{nodes." + m + "}", sink.full_name)
        #
        # return expr

    def to_df(self, dfs: dict[str, AnyFrame]) -> AnyFrame:
        """
        Execute expression on provided DataFrame `dfs`.

        Parameters
        ----------
        dfs:
            Input dataframes

        Returns
        -------
            Output dataframe
        """

        # From SQL expression
        logger.info(f"DataFrame as \n{self.expr.strip()}")

        # Read Data Sources
        for s in self.data_sources:
            dfs[f"nodes.{s.node.name}"] = s.read()

        # Convert to Native
        dfs = {k: nw.from_native(v).to_native() for k, v in dfs.items()}
        df0 = list(dfs.values())[0]

        # Get Backend
        backend = DataFrameBackends.from_df(df0)

        if backend == DataFrameBackends.POLARS:
            import polars as pl

            #
            # kwargs = {"df": df}
            # for source in self.data_sources:
            #     kwargs[f"nodes__{source.node.name}"] = source.read()
            # return pl.SQLContext(frames=dfs).execute(";".join(self.parsed_expr()))

            # Because Polars don't support {} in frame names, we use
            # double underscores (__) instead
            _dfs = {}
            for k, v in dfs.items():
                _k = "{" + k + "}"
                _dfs[to_safe_expr(_k, df_names=[k])] = v

            expr = to_safe_expr(self.expr, df_names=list(dfs.keys()))

            df = pl.SQLContext(frames=_dfs).execute(expr)
            return nw.from_native(df)

        elif backend == DataFrameBackends.PYSPARK:
            _spark = df0.sparkSession

            # Create views
            # TODO: Using parametrized queries would be ideal, but it is not compatible
            #       with older versions of spark or Delta Live Tables.
            # Because PySpark does not support view names with {}, we replaced them
            # with double underscores (__)
            for k, _df in dfs.items():
                _k = "{" + k + "}"
                _df.createOrReplaceTempView(to_safe_expr(_k, df_names=[k]))

            # Run query
            _df = None
            for expr in self.expr.split(";"):
                if expr.replace("\n", " ").strip() == "":
                    continue
                _df = _spark.sql(to_safe_expr(expr, df_names=list(dfs.keys())))
            if _df is None:
                raise ValueError(f"SQL Expression '{self.expr}' is invalid")
            return nw.from_native(_df)

        else:
            raise NotImplementedError(f"Backend '{backend}' is not supported.")



from typing import Literal

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class WorkspaceBinding(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Workspace Binding

    A binding of a workspace to some Databricks resource, such as catalog.


    Examples
    --------
    ```py
    ```
    """

    binding_type: Literal["BINDING_TYPE_READ_ONLY", "BINDING_TYPE_READ_WRITE"] = Field(
        None,
        description="Binding mode. Default to `BINDING_TYPE_READ_WRITE`. Possible values are `BINDING_TYPE_READ_ONLY`, `BINDING_TYPE_READ_WRITE`",
    )
    securable_name: str = Field(None, description="Name of securable.")
    securable_type: Literal[
        "catalog", "external_location", "storage_credential", "credential"
    ] = Field(
        None,
        description="Type of securable. Can be `catalog`, `external_location`, `storage_credential` or `credential`. Default to `catalog`",
    )
    workspace_id: int | str = Field(
        ...,
        description="The ID of the workspace to bind the resource to. Changes forces new resource.",
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self):
        return f"{self.securable_name}-{self.workspace_id}"

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:WorkspaceBinding"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_workspace_binding"



from pydantic import Field

from laktory.models.resources.providers.baseprovider import BaseProvider
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class AzureProvider(BaseProvider, PulumiResource, TerraformResource):
    """
    Azure Provider

    Examples
    --------
    ```py
    from laktory import models

    p = models.AzureProvider(
        client_id="${vars.AZURE_CLIENT_ID}",
        client_secret="${vars.AZURE_CLIENT_SECRET}",
    )
    ```
    """

    auxiliary_tenant_ids: list[str] = Field(None, description="")
    client_certificate: str = Field(
        None,
        description="""
    Base64 encoded PKCS#12 certificate bundle to use when authenticating as a Service Principal using a Client 
    Certificate
    """,
    )
    client_certificate_password: str = Field(
        None,
        description="""
    The password associated with the Client Certificate. For use when authenticating as a Service Principal using a 
    Client Certificate
    """,
    )
    client_certificate_path: str = Field(
        None,
        description="""
    The path to the Client Certificate associated with the Service Principal for use when authenticating as a Service
    Principal using a Client Certificate.
    """,
    )
    client_id: str = Field(None, description="The Client ID which should be used.")
    client_id_file_path: str = Field(
        None,
        description="The path to a file containing the Client ID which should be used.",
    )
    client_secret: str = Field(
        None,
        description="""
     The Client Secret which should be used. For use When authenticating as a Service Principal using a Client Secret.
    """,
    )
    client_secret_file_path: str = Field(
        None,
        description="""
    The path to a file containing the Client Secret which should be used. For use When authenticating as a Service 
    Principal using a Client Secret.
    """,
    )
    disable_correlation_request_id: bool = Field(
        None, description="This will disable the x-ms-correlation-request-id header."
    )
    disable_terraform_partner_id: bool = Field(
        None,
        description="""
    This will disable the Terraform Partner ID which is used if a custom partner_id isn't specified.
    """,
    )
    environment: str = Field(
        None,
        description="""
    The Cloud Environment which should be used. Possible values are public, usgovernment, and china. Defaults to public. 
    It can also be sourced from the following environment variables: AZURE_ENVIRONMENT, ARM_ENVIRONMENT
    """,
    )
    # features: ProviderFeaturesArgs = Field(None, description="")
    metadata_host: str = Field(
        None,
        description="""
    The Hostname which should be used for the Azure Metadata Service. It can also be sourced from the following 
    environment variable: ARM_METADATA_HOSTNAME
    """,
    )
    msi_endpoint: str = Field(
        None,
        description="""
    The path to a custom endpoint for Managed Service Identity - in most circumstances this should be detected
    automatically.
    """,
    )
    oidc_request_token: str = Field(
        None,
        description="""
    The bearer token for the request to the OIDC provider. For use when authenticating as a Service Principal using 
    OpenID Connect.
    """,
    )
    oidc_request_url: str = Field(
        None,
        description="""
    The URL for the OIDC provider from which to request an ID token. For use when authenticating as a Service Principal 
    using OpenID Connect.
    """,
    )
    oidc_token: str = Field(
        None,
        description="""
    The OIDC ID token for use when authenticating as a Service Principal using OpenID Connect.
    """,
    )
    oidc_token_file_path: str = Field(
        None,
        description="""
    The path to a file containing an OIDC ID token for use when authenticating as a Service Principal using OpenID 
    Connect.
    """,
    )
    partner_id: str = Field(
        None,
        description="A GUID/UUID that is registered with Microsoft to facilitate partner resource usage attribution.",
    )
    skip_provider_registration: bool = Field(
        None,
        description="""
    Should the AzureRM Provider skip registering all of the Resource Providers that it supports, if they're not already
    registered? It can also be sourced from the following environment variable: ARM_SKIP_PROVIDER_REGISTRATION
    """,
    )
    storage_use_azuread: bool = Field(
        None,
        description="""
    Should the AzureRM Provider use AzureAD to access the Storage Data Plane API's? It can also be sourced from the
    following environment variable: ARM_STORAGE_USE_AZUREAD
    """,
    )
    subscription_id: str = Field(
        None,
        description="""
    The Subscription ID which should be used. It can also be sourced from the following environment variable: 
    ARM_SUBSCRIPTION_ID
    """,
    )
    tenant_id: str = Field(None, description="The Tenant ID which should be used.")
    use_aks_workload_identity: bool = Field(
        None,
        description="Allow Azure AKS Workload Identity to be used for Authentication.",
    )
    use_cli: bool = Field(
        None, description="Allow Azure CLI to be used for Authentication."
    )
    use_msi: bool = Field(
        None,
        description="Allow Managed Service Identity to be used for Authentication.",
    )
    use_oidc: bool = Field(
        None, description="Allow OpenID Connect to be used for authentication"
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    # @property
    # def resource_key(self) -> str:
    #     return self.display_name

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "pulumi:providers:azure"



from pydantic import Field

from laktory.models.resources.databricks.storagecredential import StorageCredential
from laktory.models.resources.pulumiresource import PulumiResource


class MetastoreDataAccess(StorageCredential):
    """
    Databricks Metastore Data Access

    Examples
    --------
    ```py
    ```
    """

    is_default: bool = Field(
        None,
        description="Whether to set this credential as the default for the metastore.",
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - metastore data access credential grants
        """
        resources = []

        # Metastore data access grants
        resources += self.get_grants_additional_resources(
            object={"storage_credential": f"${{resources.{self.resource_name}.name}}"}
        )
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:MetastoreDataAccess"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_metastore_data_access"



import os
import re
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

import networkx as nx
from pydantic import AliasChoices
from pydantic import Field
from pydantic import computed_field
from pydantic import field_serializer
from pydantic import model_validator

from laktory._logger import get_logger
from laktory._settings import settings
from laktory.models import UnityCatalogDataSink
from laktory.models.basemodel import BaseModel
from laktory.models.dataquality.check import DataQualityCheck
from laktory.models.pipeline._execute import _execute  # noqa: F401
from laktory.models.pipeline._post_execute import _post_execute  # noqa: F401
from laktory.models.pipeline.orchestrators.databricksjoborchestrator import (
    DatabricksJobOrchestrator,
)
from laktory.models.pipeline.orchestrators.databrickspipelineorchestrator import (
    DatabricksPipelineOrchestrator,
)
from laktory.models.pipeline.pipelinenode import PipelineNode
from laktory.models.pipelinechild import PipelineChild
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource
from laktory.typing import AnyFrame

if TYPE_CHECKING:
    from databricks.sdk import WorkspaceClient
    from plotly.graph_objs import Figure

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Helper Functions                                                            #
# --------------------------------------------------------------------------- #


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


def parse_requirement_name(req: str) -> str | None:
    """
    Extract the package name from a requirement string.
    Returns None if it looks like a non-standard format (e.g., git+).
    """

    # Wheel file
    if req.endswith(".whl") or req.endswith(".wheel"):
        # Match the wheel filename pattern: {name}-{version}-{build?}-{python_tag}-{abi_tag}-{platform_tag}.whl
        # https://peps.python.org/pep-0427/
        filename = os.path.basename(req)
        match = re.match(
            r"^(?P<name>.+?)-\d+\.\d+(?:\.\d+)?(?:[a-zA-Z0-9]*)?-.*\.whl$", filename
        )
        if match:
            return match.group("name")
        else:
            return None

    # Git
    if req.startswith("git+") or "://" in req:
        return None

    # PyPi
    # Extract up to the first occurrence of one of these: [<=>~]
    match = re.match(r"^\s*([A-Za-z0-9_.-]+)", req)
    if match:
        return match.group(1)
    return None


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class Pipeline(BaseModel, PulumiResource, TerraformResource, PipelineChild):
    """
    Pipeline model to manage a data pipeline including reading from data sources,
    applying data transformations and outputting to data sinks.

    A pipeline is composed of collections of `nodes`, each one defining its
    own source, transformations and optional sink. A node may be the source of
    another node.

    A pipeline may be run manually by using python or the CLI, but it may also
    be deployed and scheduled using one of the supported orchestrators, such as
    a Databricks job or Lakeflow Declarative Pipeline.

    The DataFrame backend used to run the pipeline can be configured at the pipeline
    level or at the nodes level.

    Examples
    --------
    This first example shows how to configure a simple pipeline with 2 nodes.
    Upon execution, raw data will be read from a JSON files and two DataFrames
    (bronze and silver) will be created and saved as parquet files. Notice how
    the first node is used as a data source for the second node. Polars is used
    as the DataFrame backend.

    ```py
    import io

    import laktory as lk

    pipeline_yaml = '''
        name: pl-stock-prices
        dataframe_backend: POLARS

        nodes:
        - name: brz_stock_prices
          source:
            path: ./data/stock_prices/
            format: JSONL
          sinks:
          - path: ./data/brz_stock_prices.parquet
            format: PARQUET

        - name: slv_stock_prices
          source:
            node_name: brz_stock_prices
            as_stream: false
          sinks:
          - path: ./data/slv_stock_prices.parquet
            format: PARQUET
          transformer:
            nodes:
            - expr: |
                SELECT
                  CAST(data.created_at AS TIMESTAMP) AS created_at,
                  data.symbol AS name,
                  data.symbol AS symbol,
                  data.open AS open,
                  data.close AS close,
                  data.high AS high,
                  data.low AS low,
                  data.volume AS volume
                FROM
                  {df}
            - func_name: unique
              func_kwargs:
                subset:
                  - symbol
                  - created_at
                keep:
                  any
    '''

    pl = lk.models.Pipeline.model_validate_yaml(io.StringIO(pipeline_yaml))

    # Execute pipeline
    # pl.execute()
    ```

    The next example also defines a 2 nodes pipeline, but uses PySpark as the DataFrame
    backend. It defines the configuration required to deploy it as a Databricks job. In
    this case, the sinks are writing to unity catalog tables.

    ```py
    import io

    import laktory as lk

    pipeline_yaml = '''
        name: pl-stocks-job
        dataframe_backend: PYSPARK
        orchestrator:
          type: DATABRICKS_JOB

        nodes:

        - name: brz_stock_prices
          source:
            path: dbfs:/laktory/data/stock_prices/
            as_stream: false
            format: JSONL
          sinks:
          - table_name: brz_stock_prices_job
            mode: OVERWRITE

        - name: slv_stock_prices
          expectations:
          - name: positive_price
            expr: open > 0
            action: DROP
          source:
            node_name: brz_stock_prices
            as_stream: false
          sinks:
          - table_name: slv_stock_prices_job
            mode: OVERWRITE

          transformer:
            nodes:
            - expr: |
                SELECT
                    cast(data.created_at AS TIMESTAMP) AS created_at,
                    data.symbol AS symbol,
                    data.open AS open,
                    data.close AS close
                FROM
                    {df}
            - func_name: drop_duplicates
              func_kwargs:
                subset: ["created_at", "symbol"]
              dataframe_api: NATIVE
    '''
    pl = lk.models.Pipeline.model_validate_yaml(io.StringIO(pipeline_yaml))
    ```
    References
    ----------
    * [Data Pipeline](https://www.laktory.ai/concepts/pipeline/)
    """

    dependencies: list[str] = Field(
        [],
        description="""
        List of dependencies required to run the pipeline. If Laktory is not provided, it's current version is added
        to the list.
        """,
    )
    imports: list[str] = Field(
        [],
        description="""
        List of modules to import before execution. Generally used to load Narwhals extensions. 
        Packages listed in `dependencies` are automatically included in the list of imports.
        """,
    )
    name: str = Field(..., description="Name of the pipeline")
    nodes: list[PipelineNode] = Field(
        [],
        description="""
        List of pipeline nodes. Each node defines a data source, a series of transformations and optionally a sink.
        """,
    )
    orchestrator: DatabricksJobOrchestrator | DatabricksPipelineOrchestrator = Field(
        None,
        description="""
        Orchestrator used for scheduling and executing the pipeline. The
        selected option defines which resources are to be deployed.
        Supported options are instances of classes:

        - `DatabricksJobOrchestrator`: When deployed through a Databricks Job, a task
          is created for each pipeline node and all the required dependencies
          are set automatically. If a given task (or pipeline node) uses a
          `PipelineNodeDataSource` as the source, the data will be read from
          the upstream node sink.
        - `DatabricksPipelineOrchestrator`: When orchestrated through Databricks DLT, each
          pipeline node creates a DLT table (or view, if no sink is defined).
          Behind the scenes, `PipelineNodeDataSource` leverages native `dlt`
          `read` and `read_stream` functions to defined the interdependencies
          between the tables as in a standard DLT pipeline.
        """,
        # discriminator="type",  # discriminator can't be used because BaseModel adds
        # str to Literal type to support variables
    )
    root_path_: str | Path = Field(
        None,
        description="Location of the pipeline node root used to store logs, metrics and checkpoints.",
        validation_alias=AliasChoices("root_path", "root_path_"),
        exclude=True,
    )
    databricks_quality_monitor_enabled: bool = Field(
        False,
        description="Enable Databricks Quality Monitor. When enabled, quality monitors are created for each sink configured with a quality monitor and deleted for sinks without.",
    )
    _imports_imported: bool = False

    @model_validator(mode="before")
    @classmethod
    def assign_name(cls, data: Any) -> Any:
        o = data.get("orchestrator", None)
        if o and isinstance(o, dict):
            # orchestrator as a dict
            o["name"] = o.get("name", None) or data.get("name", None)
        elif isinstance(o, (DatabricksPipelineOrchestrator, DatabricksJobOrchestrator)):
            # orchestrator as a model
            o.name = o.name or o.get("name", None)

        return data

    @model_validator(mode="after")
    def validate_sinks(self) -> Any:
        from laktory.models.datasinks.hivemetastoredatasink import HiveMetastoreDataSink
        from laktory.models.datasinks.pipelineviewdatasink import PipelineViewDataSink
        from laktory.models.datasinks.unitycatalogdatasink import UnityCatalogDataSink

        o = self.orchestrator

        for n in self.nodes:
            if isinstance(o, DatabricksPipelineOrchestrator):
                for i, s in enumerate(n.sinks):
                    if not isinstance(
                        s,
                        (
                            PipelineViewDataSink,
                            HiveMetastoreDataSink,
                            UnityCatalogDataSink,
                        ),
                    ):
                        raise ValueError(
                            f"Node '{n.name}' sinks[{i}] ({type(s)}) is not supported with DLT orchestrator"
                        )

        return self

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def children_names(self):
        return ["nodes", "orchestrator"]

    # ----------------------------------------------------------------------- #
    # ID                                                                      #
    # ----------------------------------------------------------------------- #

    @property
    def resolved_name(self) -> str:
        return self.inject_vars_into_dump({"name": self.name})["name"]

    @property
    def safe_name(self):
        name = self.resolved_name

        # Replace special characters
        chars = [" ", ".", "@", "{", "}", "[", "]", "$", "|"]
        for c in chars:
            name = name.replace(c, "-")

        return name

    # ----------------------------------------------------------------------- #
    # Libraries                                                               #
    # ----------------------------------------------------------------------- #

    @property
    def _dependencies(self):
        from laktory import __version__

        laktory_found = False

        dependencies = [d for d in self.dependencies]
        for d in dependencies:
            if "laktory" in d:
                laktory_found = True

        if not laktory_found:
            dependencies += [f"laktory=={__version__}"]

        return dependencies

    @property
    def _imports(self) -> list[str]:
        pkg_names = self.imports
        for req in self.dependencies:
            name = parse_requirement_name(req)
            if name is None:
                logger.info(f"Skipping import of non-parseable requirement: {req}")
                continue
            if name not in pkg_names:
                pkg_names += [name]

        return pkg_names

    # ----------------------------------------------------------------------- #
    # Orchestrator                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def is_orchestrator_dlt(self) -> bool:
        """If `True`, pipeline orchestrator is DLT"""
        return isinstance(self.orchestrator, DatabricksPipelineOrchestrator)

    # ----------------------------------------------------------------------- #
    # Paths                                                                   #
    # ----------------------------------------------------------------------- #

    @computed_field(description="root_path")
    @property
    def root_path(self) -> Path:
        if self.root_path_:
            return Path(self.root_path_)

        return Path(settings.laktory_root) / "pipelines" / self.safe_name

    @field_serializer("root_path", when_used="json")
    def serialize_path(self, value: Path) -> str:
        return value.as_posix()

    # ----------------------------------------------------------------------- #
    # Expectations                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def checks(self) -> list[DataQualityCheck]:
        checks = []
        for node in self.nodes:
            checks += node.checks
        return checks

    # ----------------------------------------------------------------------- #
    # Nodes                                                                   #
    # ----------------------------------------------------------------------- #

    @property
    def nodes_dict(self) -> dict[str, PipelineNode]:
        """
        Nodes dictionary whose keys are the node names.

        Returns
        -------
        :
            Nodes
        """
        return {n.name: n for n in self.nodes}

    @property
    def dag(self) -> nx.DiGraph:
        """
        Networkx Directed Acyclic Graph representation of the pipeline. Useful
        to identify interdependencies between nodes.

        Returns
        -------
        :
            Directed Acyclic Graph
        """
        import networkx as nx

        # TODO: Review if this needs to be computed dynamically or if we can
        #       compute it only after initialization. It depends if we believe
        #       that the nodes will be changed after initialization.

        dag = nx.DiGraph()

        # Build nodes
        for n in self.nodes:
            dag.add_node(n.name)
        # Build edges and assign nodes to pipeline node data sources
        node_names = []
        for n in self.nodes:
            if n.name in node_names:
                raise ValueError(
                    f"Pipeline node '{n.name}' is declared twice in pipeline '{self.name}'"
                )
            node_names += [n.name]

            # for s in n.get_sources(PipelineNodeDataSource):
            for _node_name in n.upstream_node_names:
                dag.add_edge(_node_name, n.name)
                if _node_name not in self.nodes_dict:
                    raise ValueError(
                        f"Pipeline node data source '{_node_name}' is not defined in pipeline '{self.name}'"
                    )

        if not nx.is_directed_acyclic_graph(dag):
            for n in dag.nodes:
                logger.info(f"Pipeline {self.name} node: {n}")
            for e in dag.edges:
                logger.info(f"Pipeline {self.name} edge: {e[0]} -> {e[1]}")
            raise ValueError(
                f"Pipeline '{self.name}' is not a DAG (directed acyclic graph)."
                " A circular dependency has been detected. Please review nodes dependencies."
            )

        return dag

    @property
    def sorted_node_names(self):
        import networkx as nx

        return list(nx.topological_sort(self.dag))

    @property
    def sorted_nodes(self) -> list[PipelineNode]:
        """
        Topologically sorted nodes.

        Returns
        -------
        :
            List of Topologically sorted nodes.
        """
        node_names = self.sorted_node_names
        nodes = []
        for node_name in node_names:
            for node in self.nodes:
                if node.name == node_name:
                    nodes += [node]
                    break

        return nodes

    # ----------------------------------------------------------------------- #
    # Data Sources                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def data_sources(self):
        sources = []
        for n in self.nodes:
            sources += n.data_sources
        return sources

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def purge(self, spark=None) -> None:
        logger.info("Purging Pipeline")

        for inode, node in enumerate(self.sorted_nodes):
            node.purge(
                spark=spark,
            )

    def execute(
        self,
        write_sinks=True,
        full_refresh: bool = False,
        named_dfs: dict[str, AnyFrame] = None,
        update_tables_metadata: bool = True,
    ) -> None:
        """
        Execute the pipeline (read sources and write sinks) by sequentially
        executing each node. The selected orchestrator might impact how
        data sources or sinks are processed.

        Parameters
        ----------
        write_sinks:
            If `False` writing of node sinks will be skipped
        full_refresh:
            If `True` all nodes will be completely re-processed by deleting
            existing data and checkpoints before processing.
        named_dfs:
            Named DataFrames to be passed to pipeline nodes transformer.
        update_tables_metadata:
            Update tables metadata
        """
        logger.info("Executing Pipeline")

        for inode, node in enumerate(self.sorted_nodes):
            if named_dfs is None:
                named_dfs = {}

            node.execute(
                write_sinks=write_sinks,
                full_refresh=full_refresh,
                named_dfs=named_dfs,
                update_tables_metadata=update_tables_metadata,
            )

    def update_tables_metadata(self):
        logger.info("Updating pipeline tables metadata")

        for inode, node in enumerate(self.sorted_nodes):
            for s in node.sinks:
                if s.metadata:
                    s.metadata.execute()

    def update_quality_monitors(self, workspace_client: "WorkspaceClient" = None):
        if not self.databricks_quality_monitor_enabled:
            logger.info(
                f"Databricks Quality Monitor is disabled for pipeline {self.name}. Skipping update."
            )

        logger.info("Updating pipeline quality monitors")
        for inode, node in enumerate(self.sorted_nodes):
            for s in node.sinks:
                if not isinstance(s, UnityCatalogDataSink):
                    continue

                if s.databricks_quality_monitor:
                    sdk = s.databricks_quality_monitor.sdk(
                        workspace_client=workspace_client
                    )
                    sdk.create_or_update()
                else:
                    from laktory.models.resources.databricks import QualityMonitor

                    # Create a dummy quality monitor to access the delete function
                    # Current quality monitor properties will be used to delete
                    # existing assets.
                    qm = QualityMonitor(
                        table_name=s.full_name,
                        output_schema_name="dummy_schema",
                        assets_dir="dummy_path",
                        snapshot={},
                    )
                    sdk = qm.sdk(workspace_client=workspace_client)
                    sdk.delete()

    def dag_figure(self) -> "Figure":
        """
        [UNDER DEVELOPMENT] Generate a figure representation of the pipeline
        DAG.

        Returns
        -------
        :
            Plotly figure representation of the pipeline.
        """
        import networkx as nx
        import plotly.graph_objs as go

        dag = self.dag
        pos = nx.spring_layout(dag)

        # ------------------------------------------------------------------- #
        # Edges                                                               #
        # ------------------------------------------------------------------- #

        edge_traces = []
        for e in dag.edges:
            edge_traces += [
                go.Scatter(
                    x=[pos[e[0]][0], pos[e[1]][0]],
                    y=[pos[e[0]][1], pos[e[1]][1]],
                    line={
                        "color": "#a006b1",
                    },
                    marker=dict(
                        symbol="arrow-bar-up",
                        angleref="previous",
                        size=30,
                    ),
                    mode="lines+markers",
                    hoverinfo="none",
                    showlegend=False,
                )
            ]

        # ------------------------------------------------------------------- #
        # Nodes                                                               #
        # ------------------------------------------------------------------- #

        nodes_trace = go.Scatter(
            x=[_p[0] for _p in pos.values()],
            y=[_p[1] for _p in pos.values()],
            text=list(dag.nodes),
            name="pl nodes",
            mode="markers+text",
            marker=dict(
                size=50,
                color="#06d49e",
                line=dict(
                    width=2,
                    color="#dff2ed",
                ),
            ),
            textfont=dict(
                color="#a006b1",
            ),
        )

        return go.Figure(data=[nodes_trace] + edge_traces)

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def self_as_core_resources(self):
        return False

    @property
    def resource_type_id(self) -> str:
        """
        pl
        """
        return "pl"

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        if orchestrator is `DLT`:

        - DLT Pipeline

        if orchestrator is `DATABRICKS_JOB`:

        - Databricks Job
        """

        resources = []

        if self.orchestrator:
            resources += [self.orchestrator]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return ""

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return ""



from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.grants.externallocationgrant import ExternalLocationGrant
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class ExternalLocationEncryptionDetailsSseEncryptionDetails(BaseModel):
    algorith: str = Field(None, description="")
    aws_kms_key_arn: str = Field(None, description="")


class ExternalLocationEncryptionDetails(BaseModel):
    sse_encryption_details: ExternalLocationEncryptionDetailsSseEncryptionDetails = (
        Field(None, description="")
    )


class ExternalLocation(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks External Location

    Examples
    --------
    ```py
    ```
    """

    access_point: str = Field(
        None,
        description="The ARN of the s3 access point to use with the external location (AWS).",
    )
    comment: str = Field(None, description="User-supplied free-form text.")
    credential_name: str = Field(
        None,
        description="Name of the databricks.StorageCredential to use with this external location.",
    )
    encryption_details: ExternalLocationEncryptionDetails = Field(
        None,
        description="""
    The options for Server-Side Encryption to be used by each Databricks s3 client when connecting to S3 cloud 
    storage (AWS).
    """,
    )
    force_destroy: bool = Field(
        None, description="Destroy external location regardless of its dependents."
    )
    force_update: bool = Field(
        None, description="Update external location regardless of its dependents."
    )
    grant: Union[ExternalLocationGrant, list[ExternalLocationGrant]] = Field(
        None,
        description="""
    Grant(s) operating on the External Location and authoritative for a specific principal.
    Other principals within the grants are preserved. Mutually exclusive with `grants`.
    """,
    )
    grants: list[ExternalLocationGrant] = Field(
        None,
        description="""
    Grants operating on the External Location and authoritative for all principals. Replaces any existing grants 
    defined inside or outside of Laktory. Mutually exclusive with `grant`.
    """,
    )
    metastore_id: str = Field(None, description="Metastore ID")
    name: str = Field(
        None,
        description="""
    Name of External Location, which must be unique within the databricks_metastore. Change forces creation of a new
    resource.
    """,
    )
    owner: str = Field(
        None,
        description="Username/groupname/sp application_id of the external location owner.",
    )
    read_only: bool = Field(
        None, description="Indicates whether the external location is read-only."
    )
    skip_validation: bool = Field(
        None,
        description="Suppress validation errors if any & force save the external location",
    )
    url: str = Field(
        None,
        description="""
    Path URL in cloud storage, of the form: s3://[bucket-host]/[bucket-dir] (AWS), abfss://[user]@[host]/[path]
    (Azure), gs://[bucket-host]/[bucket-dir] (GCP).
    """,
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #
    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - external location grants
        """
        resources = []

        # External Location Grants
        resources += self.get_grants_additional_resources(
            object={"external_location": f"${{resources.{self.resource_name}.id}}"}
        )
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:ExternalLocation"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["grant", "grants"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_external_location"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from pydantic import Field

from laktory._logger import get_logger
from laktory.models.datasources.basedatasource import BaseDataSource
from laktory.typing import AnyFrame

logger = get_logger(__name__)


class PipelineNodeDataSource(BaseDataSource):
    """
    Data Source using an upstream pipeline node. Using a pipeline node data
    source defines the interdependencies between each node in a pipeline.
    Depending on the selected pipeline orchestrator and the context, a pipeline
    node data source might read the data from:

    - memory
    - upstream node sink
    - DLT table

    Examples
    ---------
    ```python
    from laktory import models

    brz = models.PipelineNode(
        name="brz_stock_prices",
        source={
            "path": "/Volumes/sources/landing/events/yahoo-finance/stock_price",
            "format": "JSON",
        },
        sinks=[
            {
                "path": "/Volumes/sources/landing/tables/brz_stock_prices",
                "format": "PARQUET",
            }
        ],
    )

    slv = models.PipelineNode(
        name="slv_stock_prices",
        source={"node_name": "brz_stock_prices"},
        sinks=[
            {
                "path": "/Volumes/sources/landing/tables/slv_stock_prices",
                "format": "PARQUET",
            }
        ],
    )

    pl = models.Pipeline(name="pl-stock-prices", nodes=[brz, slv])

    # pl.execute(spark=spark)
    ```
    References
    ----------
    * [Data Sources and Sinks](https://www.laktory.ai/concepts/sourcessinks/)
    """

    node_name: str = Field(..., description="Name of the upstream pipeline node")
    type: str = Field("PIPELINE_NODE", frozen=True)

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self) -> str:
        return self.node_name

    @property
    def node(self):
        pl = self.parent_pipeline

        if pl is None:
            raise ValueError(f"Source '{self.node_name}' is not attached to a pipeline")

        if self.node_name not in pl.sorted_node_names:
            raise ValueError(
                f"Node '{self.node_name}' does not exists in pipeline '{pl.name}'"
            )

        return pl.nodes_dict[self.node_name]

    @property
    def sink_table_full_name(self):
        from laktory.models.datasinks.tabledatasink import TableDataSink

        node = self.node
        if not node.primary_sink:
            raise ValueError(
                f"Source node '{self.node_name}' doest not have a sink defined"
            )
        if not isinstance(node.primary_sink, TableDataSink):
            raise ValueError(
                f"Source node '{self.node_name}' sink is not of type `TableDataSink`"
            )
        return node.primary_sink.full_name

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read_spark(self) -> AnyFrame:
        stream_to_batch = not self.as_stream and self.node.source.as_stream
        is_dlt = False

        pl = self.parent_pipeline
        is_orchestrator_dlt = pl is not None and pl.is_orchestrator_dlt

        if is_orchestrator_dlt:
            from laktory import is_dlt_execute

            is_dlt = is_dlt_execute()

        # Reading from DLT
        if is_dlt:
            import dlt

            if self.as_stream:
                logger.info(f"Reading pipeline node {self._id} with DLT as stream")
                df = dlt.read_stream(self.node.primary_sink.dlt_table_or_view_name)
            else:
                logger.info(f"Reading pipeline node {self._id} with DLT as static")
                df = dlt.read(self.node.primary_sink.dlt_table_or_view_name)

        elif stream_to_batch or self.node.output_df is None:
            if self.node.has_sinks:
                logger.info(f"Reading pipeline node {self._id} from primary sink")
                df = self.node.primary_sink.read(as_stream=self.as_stream)
            else:
                logger.info("Executing parent pipeline node")
                self.node.execute()
                df = self.node.output_df

        elif self.node.output_df is not None:
            logger.info(f"Reading pipeline node {self._id} from output DataFrame")
            df = self.node.output_df

        else:
            raise ValueError(f"Pipeline Node {self._id} can't read DataFrame")

        return df

    def _read_polars(self) -> AnyFrame:
        # Read from node output DataFrame (if available)
        if self.node.output_df is not None:
            logger.info(f"Reading pipeline node {self._id} from output DataFrame")
            df = self.node.output_df

        # Read from node sink
        elif self.node.primary_sink:
            logger.info(f"Reading pipeline node {self._id} from sink")
            df = self.node.primary_sink.read(as_stream=self.as_stream)

        else:
            raise ValueError(f"Pipeline Node {self._id} can't read DataFrame")

        return df



from typing import Any
from typing import Literal
from typing import Union

from pydantic import Field
from pydantic import model_validator

from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class WarehouseCustomTag(BaseModel):
    key: str = Field(..., description="Tag key")
    value: str = Field(..., description="Tag value")


class WarehouseTags(BaseModel):
    custom_tags: list[WarehouseCustomTag] = Field([], description="Tags specifications")

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "custom_tags": "custom_tags",
        }


class WarehouseLookup(ResourceLookup):
    id: str = Field(
        serialization_alias="id",
        default=None,
        description="The ID of the SQL warehouse.",
    )
    name: str = Field(
        None,
        description="""
        Name of the SQL warehouse. Name of the SQL warehouse to search (case-sensitive). Argument only supported by 
        Terraform IaC backend.
        """,
    )

    @model_validator(mode="after")
    def at_least_one(self) -> Any:
        if self.id is None and self.name is None:
            raise ValueError("At least `id` or `name` must be set.")

        if not (self.id is None or self.name is None):
            raise ValueError("Only one of `id` or `name` must be set.")

        return self


class Warehouse(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Warehouse

    Examples
    --------
    ```py
    from laktory import models

    warehouse = models.resources.databricks.Warehouse(
        name="default",
        cluster_size="2X-Small",
        auto_stop_mins=30,
        channel_name="CHANNEL_NAME_PREVIEW",
        enable_photon=True,
        enable_serverless_compute=True,
        access_controls=[
            {"group_name": "account users", "permission_level": "CAN_USE"}
        ],
    )
    ```
    """

    cluster_size: Literal[
        "2X-Small",
        "X-Small",
        "Small",
        "Medium",
        "Large",
        "X-Large",
        "2X-Large",
        "3X-Large",
        "4X-Large",
    ] = Field(..., description="The size of the clusters allocated to the endpoint")
    access_controls: list[AccessControl] = Field([], description="Access controls list")
    auto_stop_mins: int = Field(
        None,
        description="Time in minutes until an idle SQL warehouse terminates all clusters and stops.",
    )
    channel_name: Union[
        Literal["CHANNEL_NAME_CURRENT", "CHANNEL_NAME_PREVIEW"], str
    ] = Field(None, description="Channel specifications")
    # data_source_id
    enable_photon: bool = Field(None, description="If `True`, photon is enabled")
    enable_serverless_compute: bool = Field(
        None, description="If `True`, warehouse is serverless."
    )
    instance_profile_arn: str = Field(None, description="")
    jdbc_url: str = Field(None, description="JDBC connection string.")
    lookup_existing: WarehouseLookup = Field(
        None,
        exclude=True,
        description="Specifications for looking up existing resource. Other attributes will be ignored.",
    )
    max_num_clusters: int = Field(
        None,
        description="Maximum number of clusters available when a SQL warehouse is running.",
    )
    min_num_clusters: int = Field(
        None,
        description="Minimum number of clusters available when a SQL warehouse is running.",
    )
    name: str = Field(..., description="Warehouse name")
    num_clusters: int = Field(
        None, description="Fixed number of clusters when autoscaling is not enabled."
    )
    # odbc_params
    spot_instance_policy: Literal["COST_OPTIMIZED", "RELIABILITY_OPTIMIZED"] = Field(
        None, description="The spot policy to use for allocating instances to clusters."
    )
    # state
    tags: WarehouseTags = Field(
        None, description="Databricks tags all endpoint resources with these tags."
    )
    warehouse_type: Literal["CLASSIC", "PRO"] = Field(
        None, description="SQL warehouse type."
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - warehouse permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    sql_endpoint_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:SqlEndpoint"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls"]

    @property
    def pulumi_properties(self):
        d = super().pulumi_properties
        if self._camel_serialization:
            d["channel"] = {"name": d.pop("channelName", None)}
        else:
            d["channel"] = {"name": d.pop("channel_name", None)}
        return d

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "tags": "tags",
        }

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_sql_endpoint"

    @property
    def terraform_resource_lookup_type(self) -> str:
        return "databricks_sql_warehouse"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes

    @property
    def terraform_properties(self) -> dict:
        d = super().terraform_properties
        d["channel"] = {"name": d.pop("channel_name", None)}
        return d



from laktory.dispatcher.databricksjobrunner import DatabricksJobRunner
from laktory.dispatcher.databrickspipelinerunner import DatabricksPipelineRunner
from laktory.dispatcher.dispatcher import Dispatcher
from laktory.dispatcher.dispatcherrunner import DispatcherRunner



import narwhals as nw

from laktory._logger import get_logger
from laktory.typing import AnyFrame

logger = get_logger(__name__)


def window_filter(
    self,
    partition_by: str | list[str] | None,
    order_by: str | list[str] | None = None,
    drop_row_index: bool = True,
    row_index_name: str = "_row_index",
    rows_to_keep: int = 1,
) -> AnyFrame:
    """
    Apply spark window-based filtering

    Parameters
    ----------
    partition_by
        Defines the columns used for grouping into Windows
    order_by:
        Defines the column used for sorting before dropping rows
    drop_row_index:
        If `True`, the row index column is dropped
    row_index_name:
        Group-specific and sorted rows index name
    rows_to_keep:
        How many rows to keep per window


    Examples
    --------
    ```py
    import narwhals as nw
    import polars as pl

    import laktory as lk  # noqa: F401

    df0 = nw.from_native(
        pl.DataFrame(
            [
                ["2023-01-01T00:00:00Z", "APPL", 200.0],
                ["2023-01-02T00:00:00Z", "APPL", 202.0],
                ["2023-01-03T00:00:00Z", "APPL", 201.5],
                ["2023-01-01T00:00:00Z", "GOOL", 200.0],
                ["2023-01-02T00:00:00Z", "GOOL", 202.0],
                ["2023-01-03T00:00:00Z", "GOOL", 201.5],
            ],
            ["created_at", "symbol", "price"],
        ).with_columns(pl.col("created_at").cast(pl.Datetime))
    )

    df = df0.laktory.window_filter(
        partition_by="symbol",
        order_by=["created_at"],
        drop_row_index=False,
        rows_to_keep=1,
    )

    print(df)
    '''
    ┌─────────────────────────────────────────────────────┐
    |                 Narwhals DataFrame                  |
    |-----------------------------------------------------|
    || created_at          | symbol | price | _row_index ||
    ||---------------------|--------|-------|------------||
    || 2023-01-01 00:00:00 | APPL   | 200.0 | 0          ||
    || 2023-01-01 00:00:00 | GOOL   | 200.0 | 0          ||
    └─────────────────────────────────────────────────────┘
    '''
    ```
    """
    if rows_to_keep < 1:
        raise ValueError("`rows_to_keep` must be >= 1")

    # Row Index
    df = self._df.laktory.with_row_index(
        name=row_index_name, partition_by=partition_by, order_by=order_by
    )

    # Filter
    df = df.filter(nw.col(row_index_name) <= rows_to_keep - 1)
    if drop_row_index:
        df = df.drop(row_index_name)

    return df



from typing import Literal

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource


class StorageBlob(BaseModel, PulumiResource):
    """
    Databricks group
    """

    access_tier: Literal["HOT", "COOL", "ARCHIVE"] = Field(
        None, description="The access tier of the storage blob."
    )
    account_name: str = Field(
        None,
        description="Specifies the storage account in which to create the storage container.",
    )
    blob_name: str = Field(
        None,
        description="The name of the storage blob. Must be unique within the storage container the blob is located.",
    )
    container_name: str = Field(
        None,
        description="The name of the storage container in which this blob should be created.",
    )
    content_md5: str = Field(
        None,
        description="The MD5 sum of the blob contents. Cannot be defined if blob type is Append.",
    )
    content_type: str = Field(
        None,
        description="The content type of the storage blob. Defaults to application/octet-stream.",
    )
    metadata: dict[str, str] = Field(None, description="A map of custom blob metadata.")
    resource_group_name: str = Field(
        None,
        description="The name of the resource group within the user's subscription.",
    )
    # source: Union[pulumi.Asset, pulumi.Archive] = None
    type: Literal["BLOCK", "APPEND"] = Field(
        None,
        description="The type of the storage blob to be created. Defaults to 'Block'.",
    )

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """Blob name"""
        return self.blob_name

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "azure-native:storage:Blob"

    # # ----------------------------------------------------------------------- #
    # # Terraform Properties                                                    #
    # # ----------------------------------------------------------------------- #
    #
    # @property
    # def terraform_resource_type(self) -> str:
    #     return "databricks_group"
    #
    # @property
    # def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
    #     return self.pulumi_excludes
